import os
import random
import math
import torch
import torch.nn.functional as F
import numpy as np
import pandas as pd
import torch.optim as optim
from typing import Optional, Any, Union
from torch import nn
from scipy.special import softmax
from medcat.config.config_meta_cat import ConfigMetaCAT
from medcat.components.addons.meta_cat.mctokenizers.tokenizers import (
    TokenizerWrapperBase)
from sklearn.metrics import (
    classification_report, precision_recall_fscore_support, confusion_matrix)
from sklearn.model_selection import train_test_split
from sklearn.utils.class_weight import compute_class_weight
from transformers import get_linear_schedule_with_warmup
from torch.optim import AdamW


import logging

logger = logging.getLogger(__name__)


def set_all_seeds(seed: int) -> None:
    torch.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)


def create_batch_piped_data(data: list[tuple[list[int], int, Optional[int]]],
                            start_ind: int, end_ind: int,
                            device: Union[torch.device, str],
                            pad_id: int) -> tuple:
    """Creates a batch given data and start/end that denote batch size,
    will also add padding and move to the right device.

    Args:
        data (list[tuple[list[int], int, Optional[int]]]):
            Data in the format: [[<[input_ids]>, <cpos>, Optional[int]], ...],
            the third column is optional and represents the output label
        start_ind (int):
            Start index of this batch
        end_ind (int):
            End index of this batch
        device (Union[torch.device, str]):
            Where to move the data
        pad_id (int):
            Padding index

    Returns:
        x ():
            Same as data, but subsetted and as a tensor
        cpos ():
            Center positions for the data
        attention_mask:
            Indicating padding mask for the data
        y:
            class label of the data
    """
    max_seq_len = max([len(x[0]) for x in data])
    x = [x[0][0:max_seq_len] + [pad_id] * max(0, max_seq_len - len(x[0]))
         for x in data[start_ind:end_ind]]
    cpos = [x[1] for x in data[start_ind:end_ind]]
    y = None
    if len(data[0]) == 3:
        # Means we have the y column
        y = torch.tensor([x[2] for x in data[start_ind:end_ind]],
                         dtype=torch.long).to(device)

    x2 = torch.tensor(x, dtype=torch.long).to(device)
    # cpos = torch.tensor(cpos, dtype=torch.long).to(device)
    attention_masks = (x2 != pad_id).type(torch.int)
    return x2, cpos, attention_masks, y


def predict(model: nn.Module, data: list[tuple[list[int], int, Optional[int]]],
            config: ConfigMetaCAT) -> tuple:
    """Predict on data used in the meta_cat.pipe

    Args:
        model (nn.Module):
            The model.
        data (list[tuple[list[int], int, Optional[int]]]):
            Data in the format: [[<input_ids>, <cpos>], ...]
        config (ConfigMetaCAT):
            Configuration for this meta_cat instance.

    Returns:
        predictions (list[int]):
            For each row of input data a prediction
        confidence (list[float]):
            For each prediction a confidence value
    """

    pad_id = config.model.padding_idx
    batch_size = config.general.batch_size_eval
    device = config.general.device
    ignore_cpos = config.model.ignore_cpos

    model.eval()
    model.to(device)

    num_batches = math.ceil(len(data) / batch_size)
    all_logits = []

    with torch.no_grad():
        for i in range(num_batches):
            x, cpos, attention_masks, _ = create_batch_piped_data(
                data, i * batch_size, (i + 1) * batch_size,
                device=device, pad_id=pad_id)

            logits = model(x, center_positions=cpos,
                           attention_mask=attention_masks,
                           ignore_cpos=ignore_cpos)
            all_logits.append(logits.detach().cpu().numpy())

    predictions = []
    confidences = []

    # Can be that there are not logits, data is empty
    if all_logits:
        logits = np.concatenate(all_logits, axis=0)
        predictions = np.argmax(logits, axis=1)
        confidences = np.max(softmax(logits, axis=1), axis=1)

    return predictions, confidences


def split_list_train_test(data: list, test_size: float, shuffle: bool = True
                          ) -> tuple:
    """Shuffle and randomly split data

    Args:
        data (list): The data.
        test_size (float): The test size.
        shuffle (bool): Whether to shuffle the data. Defaults to True.

    Returns:
        tuple: The train data, and the test data.
    """
    if shuffle:
        random.shuffle(data)

    X_features = [x[:-1] for x in data]
    y_labels = [x[-1] for x in data]

    X_train, X_test, y_train, y_test = train_test_split(
        X_features, y_labels, test_size=test_size, random_state=42)

    train_data = [x + [y] for x, y in zip(X_train, y_train)]
    test_data = [x + [y] for x, y in zip(X_test, y_test)]

    return train_data, test_data


def print_report(epoch: int, running_loss: list, all_logits: list,
                 y: Any, name: str = 'Train') -> None:
    """Prints some basic stats during training

    Args:
        epoch (int): Number of epochs.
        running_loss (list): The loss
        all_logits (list): List of logits
        y (Any): The y array.
        name (str): The name of the report. Defaults to Train.
    """
    if all_logits:
        logger.info('Epoch: %d %s %s', epoch, "*" * 50, name)
        logger.info(classification_report(
            y, np.argmax(np.concatenate(all_logits, axis=0), axis=1)))


class FocalLoss(nn.Module):
    def __init__(self, alpha=None, gamma=2):
        super(FocalLoss, self).__init__()
        self.alpha = alpha
        self.gamma = gamma

    def forward(self, inputs, targets):
        ce_loss = F.cross_entropy(inputs, targets, reduction='none')
        pt = torch.exp(-ce_loss)
        loss = (self.alpha[targets] * (1 - pt) ** self.gamma * ce_loss).mean()
        return loss


def train_model(model: nn.Module, data: list, config: ConfigMetaCAT,
                save_dir_path: Optional[str] = None) -> dict:
    """Trains a LSTM model and BERT with autocheckpoints

    Args:
        model (nn.Module): The model
        data (list): The data.
        config (ConfigMetaCAT): MetaCAT config.
        save_dir_path (Optional[str]): The save dir path if required.
            Defaults to None.

    Returns:
        dict: The classification report for the winner.

    Raises:
        Exception: If auto-save is enabled but no save dir path is provided.
    """
    # Get train/test from data
    train_data, test_data = split_list_train_test(
        data, test_size=config.train.test_size,
        shuffle=config.train.shuffle_data)
    device = torch.device(config.general.device)  # Create a torch device

    class_weights = config.train.class_weights

    if class_weights is None:
        if config.train.compute_class_weights is True:
            y_ = [x[2] for x in train_data]
            class_weights = compute_class_weight(
                class_weight="balanced", classes=np.unique(y_), y=y_)
            config.train.class_weights = class_weights.tolist()
            logger.info("Class weights computed: %s", class_weights)

            class_weights = torch.FloatTensor(class_weights).to(device)
            if config.train.loss_funct == 'cross_entropy':
                criterion: Union[FocalLoss,
                                 nn.CrossEntropyLoss] = nn.CrossEntropyLoss(
                    weight=class_weights)
            elif config.train.loss_funct == 'focal_loss':
                criterion = FocalLoss(
                    alpha=class_weights, gamma=config.train.gamma)

        else:
            logger.warning(
                "Class weights not provided and compute_class_weights "
                "parameter is set to False. No class weights used for "
                "training.")
            if config.train.loss_funct == 'cross_entropy':
                criterion = nn.CrossEntropyLoss()
            elif config.train.loss_funct == 'focal_loss':
                criterion = FocalLoss(gamma=config.train.gamma)
    else:
        class_weights = torch.FloatTensor(class_weights).to(device)
        if config.train.loss_funct == 'cross_entropy':
            criterion = nn.CrossEntropyLoss(
                weight=class_weights)
        elif config.train.loss_funct == 'focal_loss':
            criterion = FocalLoss(
                alpha=class_weights, gamma=config.train.gamma)

    parameters = filter(lambda p: p.requires_grad, model.parameters())

    def initialize_model(classifier, data_, batch_size_, lr_, epochs=4):
        """Initialize the Classifier, the optimizer and the learning
        rate scheduler.

            Args:
                classifier (nn.Module):
                    The model to be trained
                data_ (list):
                    The data
                batch_size_:
                    Batch size
                lr_:
                    Learning rate for training
                epochs:
                    Number of training iterations

            Returns:
                classifier:
                    model
                optimizer_:
                    optimizer
                scheduler_:
                    scheduler
            """

        # Create the optimizer
        optimizer_ = AdamW(classifier.parameters(),
                           lr=lr_,  # Default learning rate
                           eps=1e-8,  # Default epsilon value
                           weight_decay=1e-5
                           )

        # Total number of training steps
        total_steps = int((len(data_) / batch_size_) * epochs)
        logger.info('Total steps for optimizer: %d', total_steps)

        # Set up the learning rate scheduler
        scheduler_ = get_linear_schedule_with_warmup(
            optimizer_, num_warmup_steps=0, num_training_steps=total_steps)
        return classifier, optimizer_, scheduler_

    batch_size = config.train.batch_size
    batch_size_eval = config.general.batch_size_eval
    pad_id = config.model.padding_idx
    nepochs = config.train.nepochs
    ignore_cpos = config.model.ignore_cpos
    num_batches = math.ceil(len(train_data) / batch_size)
    num_batches_test = math.ceil(len(test_data) / batch_size_eval)
    optimizer = optim.Adam(
        parameters, lr=config.train.lr, weight_decay=1e-5)
    if config.model.model_architecture_config is not None:
        if config.model.model_architecture_config['lr_scheduler'] is True:
            model, optimizer, scheduler = initialize_model(
                model, train_data, batch_size, config.train.lr,
                epochs=nepochs)

    model.to(device)  # Move the model to device

    # Can be pre-calculated for the whole dataset
    y_test = [x[2] for x in test_data]
    y_train = [x[2] for x in train_data]

    winner_report: dict = {}
    for epoch in range(nepochs):
        running_loss = []
        all_logits = []
        model.train()
        for i in range(num_batches):
            model.zero_grad()

            x, cpos, attention_masks, y = create_batch_piped_data(
                train_data, i * batch_size, (i + 1) * batch_size,
                device=device, pad_id=pad_id)
            logits = model(x, attention_mask=attention_masks,
                           center_positions=cpos, ignore_cpos=ignore_cpos)
            loss = criterion(logits, y)
            loss.backward()
            # Track loss and logits
            running_loss.append(loss.item())
            all_logits.append(logits.detach().cpu().numpy())

            parameters = filter(lambda p: p.requires_grad, model.parameters())
            nn.utils.clip_grad_norm_(parameters, 0.15)
            optimizer.step()
            if config.model.model_architecture_config is not None:
                lr_scheduler = config.model.model_architecture_config[
                    'lr_scheduler']
                if lr_scheduler is True:
                    scheduler.step()

        all_logits_test = []
        running_loss_test = []
        model.eval()

        with torch.no_grad():
            for i in range(num_batches_test):
                x, cpos, attention_masks, y = create_batch_piped_data(
                    test_data, i * batch_size_eval, (i + 1) * batch_size_eval,
                    device=device, pad_id=pad_id)
                logits = model(x, attention_mask=attention_masks,
                               center_positions=cpos, ignore_cpos=ignore_cpos)

                # Track loss and logits
                running_loss_test.append(loss.item())
                all_logits_test.append(logits.detach().cpu().numpy())

        print_report(epoch, running_loss, all_logits, y=y_train, name='Train')
        print_report(epoch, running_loss_test, all_logits_test, y=y_test,
                     name='Test')

        _report = classification_report(y_test, np.argmax(np.concatenate(
            all_logits_test, axis=0), axis=1), output_dict=True,
            zero_division=0)
        if not winner_report or _report[config.train.metric['base']][
            config.train.metric['score']] > \
                winner_report['report'][config.train.metric['base']][
                    config.train.metric['score']]:

            report = classification_report(y_test, np.argmax(np.concatenate(
                all_logits_test, axis=0), axis=1), output_dict=True)
            cm = confusion_matrix(y_test, np.argmax(np.concatenate(
                all_logits_test, axis=0), axis=1), normalize='true')
            report_train = classification_report(y_train, np.argmax(
                np.concatenate(all_logits, axis=0), axis=1), output_dict=True,
                zero_division=0)

            winner_report['confusion_matrix'] = cm
            winner_report['report'] = report
            winner_report['report_train'] = report_train
            winner_report['epoch'] = epoch

            # Save if needed
            if config.train.auto_save_model:
                if save_dir_path is None:
                    raise Exception(
                        "The `save_dir_path` argument is required if "
                        "`aut_save_model` is `True` in the config")
                else:
                    path = os.path.join(save_dir_path, 'model.dat')
                    torch.save(model.state_dict(), path)
                    logger.info("\n##### Model saved to %s at epoch: %d and "
                                "%s/%s: %s #####\n", path, epoch,
                                config.train.metric['base'],
                                config.train.metric['score'],
                                winner_report['report'][config.train.metric[
                                    'base']][config.train.metric['score']])

    return winner_report


def eval_model(model: nn.Module, data: list, config: ConfigMetaCAT,
               tokenizer: TokenizerWrapperBase) -> dict:
    """Evaluate a trained model on the provided data

    Args:
        model (nn.Module): The model.
        data (list): The data.
        config (ConfigMetaCAT): The MetaCAT config.
        tokenizer (TokenizerWrapperBase): The tokenizer.

    Returns:
        dict: Results (precision, recall, f1, examples, confusion matrix)
    """
    device = torch.device(config.general.device)  # Create a torch device
    batch_size_eval = config.general.batch_size_eval
    pad_id = config.model.padding_idx
    ignore_cpos = config.model.ignore_cpos
    class_weights = config.train.class_weights

    if class_weights is not None:
        class_weights = torch.FloatTensor(class_weights).to(device)
        # Set the criterion to Cross Entropy Loss
        criterion = nn.CrossEntropyLoss(weight=class_weights)
    else:
        # Set the criterion to Cross Entropy Loss
        criterion = nn.CrossEntropyLoss()

    y_eval = [x[2] for x in data]
    num_batches = math.ceil(len(data) / batch_size_eval)
    running_loss = []
    all_logits = []
    model.to(device)
    model.eval()

    with torch.no_grad():
        for i in range(num_batches):
            x, cpos, attention_masks, y = create_batch_piped_data(
                data, i * batch_size_eval, (i + 1) * batch_size_eval,
                device=device, pad_id=pad_id)

            logits = model(x, center_positions=cpos,
                           attention_mask=attention_masks,
                           ignore_cpos=ignore_cpos)

            loss = criterion(logits, y)

            # Track loss and logits
            running_loss.append(loss.item())
            all_logits.append(logits.detach().cpu().numpy())

    print_report(0, running_loss, all_logits, y=y_eval, name='Eval')

    score_average = config.train.score_average
    predictions = np.argmax(np.concatenate(all_logits, axis=0), axis=1)
    precision, recall, f1, support = precision_recall_fscore_support(
        y_eval, predictions, average=score_average)

    labels = [name for (name, _) in sorted(
        config.general.category_value2id.items(), key=lambda x: x[1])]
    labels_present_: set = set(predictions)
    labels_present: list[str] = [str(x) for x in labels_present_]

    if len(labels) != len(labels_present):
        logger.warning(
            "The evaluation dataset does not contain all the labels, some "
            "labels are missing. Performance displayed for labels found...")
    confusion = pd.DataFrame(
        data=confusion_matrix(y_eval, predictions, ),
        columns=["true " + label for label in labels_present],
        index=["predicted " + label for label in labels_present],
    )

    examples: dict = {'FP': {}, 'FN': {}, 'TP': {}}
    id2category_value = {v: k for k, v
                         in config.general.category_value2id.items()}
    for i, p in enumerate(predictions):
        y = id2category_value[y_eval[i]]
        p = id2category_value[p]
        c = data[i][1]
        if isinstance(c, list):
            c = c[-1]

        tkns = data[i][0]
        assert tokenizer.hf_tokenizers is not None
        text = (tokenizer.hf_tokenizers.decode(tkns[0:c]) + " <<" +
                tokenizer.hf_tokenizers.decode(
                    tkns[c:c + 1]).strip() + ">> " +
                tokenizer.hf_tokenizers.decode(tkns[c + 1:]))
        info = "Predicted: {}, True: {}".format(p, y)
        if p != y:
            # We made a mistake
            examples['FN'][y] = examples['FN'].get(y, []) + [(info, text)]
            examples['FP'][p] = examples['FP'].get(p, []) + [(info, text)]
        else:
            examples['TP'][y] = examples['TP'].get(y, []) + [(info, text)]

    return {'precision': precision, 'recall': recall, 'f1': f1,
            'examples': examples, 'confusion matrix': confusion}



from typing import Optional, Iterable, Union
import logging
import os

from medcat.utils.defaults import COMPONENTS_FOLDER
from medcat.tokenizing.tokenizers import BaseTokenizer, create_tokenizer
from medcat.components.types import (
    CoreComponentType, create_core_component, CoreComponent, BaseComponent,
    AbstractCoreComponent)
from medcat.components.addons.addons import AddonComponent, create_addon
from medcat.tokenizing.tokens import (MutableDocument, MutableEntity,
                                      MutableToken)
from medcat.storage.serialisers import (
    AvailableSerialisers, serialise, deserialise)
from medcat.storage.serialisables import Serialisable
from medcat.vocab import Vocab
from medcat.cdb import CDB
from medcat.config import Config
from medcat.config.config import ComponentConfig
from medcat.config.config_meta_cat import ConfigMetaCAT
from medcat.config.config_rel_cat import ConfigRelCAT


logger = logging.getLogger(__name__)


class DelegatingTokenizer(BaseTokenizer):
    """A delegating tokenizer.

    This can be used to create a tokenizer with some preprocessing
    (i.e components) included.
    """

    def __init__(self, tokenizer: BaseTokenizer,
                 components: list[CoreComponent]):
        self.tokenizer = tokenizer
        self.components = components

    def create_entity(self, doc: MutableDocument,
                      token_start_index: int, token_end_index: int,
                      label: str) -> MutableEntity:
        return self.tokenizer.create_entity(
            doc, token_start_index, token_end_index, label)

    def entity_from_tokens(self, tokens: list[MutableToken]) -> MutableEntity:
        return self.tokenizer.entity_from_tokens(tokens)

    def __call__(self, text: str) -> MutableDocument:
        doc = self.tokenizer(text)
        for comp in self.components:
            doc = comp(doc)
        return doc

    @classmethod
    def create_new_tokenizer(cls, config: Config) -> 'DelegatingTokenizer':
        raise ValueError("Initialise the delegating tokenizer with its initialiser")

    def get_doc_class(self) -> type[MutableDocument]:
        return self.tokenizer.get_doc_class()

    def get_entity_class(self) -> type[MutableEntity]:
        return self.tokenizer.get_entity_class()


class Pipeline:
    """The pipeline for the NLP process.

    This class is responsible to initial creation of the NLP document,
    as well as running through of all the components and addons.
    """

    def __init__(self, cdb: CDB, vocab: Optional[Vocab],
                 model_load_path: Optional[str],
                 # NOTE: upon reload, old pipe can be useful
                 old_pipe: Optional['Pipeline'] = None):
        self.cdb = cdb
        # NOTE: Vocab is None in case of DeID models and thats fine then,
        #       but it should be non-None otherwise
        self.vocab: Vocab = vocab  # type: ignore
        self.config = self.cdb.config
        self._tokenizer = self._init_tokenizer()
        self._components: list[CoreComponent] = []
        self._addons: list[AddonComponent] = []
        self._init_components(model_load_path, old_pipe)

    @property
    def tokenizer(self) -> BaseTokenizer:
        """The raw tokenizer (with no components)."""
        return self._tokenizer

    @property
    def tokenizer_with_tag(self) -> BaseTokenizer:
        """The tokenizer with the tagging component."""
        tag_comp = self.get_component(CoreComponentType.tagging)
        return DelegatingTokenizer(self.tokenizer, [tag_comp])

    def _init_tokenizer(self) -> BaseTokenizer:
        nlp_cnf = self.config.general.nlp
        try:
            return create_tokenizer(nlp_cnf.provider, self.config)
        except TypeError as type_error:
            if nlp_cnf.provider == 'spacy':
                raise type_error
            raise IncorrectArgumentsForTokenizer(
                nlp_cnf.provider) from type_error

    def _init_component(self, comp_type: CoreComponentType,
                        model_load_path: Optional[str]) -> CoreComponent:
        comp_config: ComponentConfig = getattr(self.config.components,
                                               comp_type.name)
        comp_name = comp_config.comp_name
        try:
            comp = create_core_component(
                comp_type, comp_name, comp_config, self.tokenizer, self.cdb,
                self.vocab, model_load_path)
        except TypeError as type_error:
            if comp_name == 'default':
                raise type_error
            raise IncorrectArgumentsForComponent(
                comp_type, comp_name) from type_error
        return comp

    def _get_loaded_components_paths(self, model_load_path: Optional[str]
                                     ) -> tuple[dict[str, str],
                                                dict[tuple[str, str], str]]:
        loaded_core_component_paths: dict[str, str] = {}
        loaded_addon_component_paths: dict[tuple[str, str], str] = {}
        if not model_load_path:
            return loaded_core_component_paths, loaded_addon_component_paths
        components_folder = os.path.join(
            model_load_path, COMPONENTS_FOLDER)
        if not os.path.exists(components_folder):
            return loaded_core_component_paths, loaded_addon_component_paths
        for folder_name in os.listdir(components_folder):
            cur_folder_path = os.path.join(
                components_folder, folder_name)
            if folder_name.startswith(AbstractCoreComponent.NAME_PREFIX):
                loaded_core_component_paths[
                    folder_name[len(AbstractCoreComponent.NAME_PREFIX):]
                    ] = cur_folder_path
            elif folder_name.startswith(AddonComponent.NAME_PREFIX):
                addon_folder_name = folder_name[
                    len(AddonComponent.NAME_PREFIX):]
                addon_type, addon_name = addon_folder_name.split(
                    AddonComponent.NAME_SPLITTER, 1)
                loaded_addon_component_paths[
                    (addon_type, addon_name)] = cur_folder_path
            else:
                raise ValueError()
        return loaded_core_component_paths, loaded_addon_component_paths

    def _load_saved_core_component(self, cct_name: str, comp_folder_path: str
                                   ) -> CoreComponent:
        logger.info("Using loaded component for '%s' for", cct_name)
        cnf: ComponentConfig = getattr(self.config.components, cct_name)
        comp = deserialise(
            comp_folder_path,
            # NOTE: the following are keyword arguments used
            #       for manual deserialisation
            cnf=cnf, tokenizer=self.tokenizer, cdb=self.cdb,
            vocab=self.vocab, model_load_path=os.path.dirname(
                os.path.dirname(comp_folder_path)))
        if not isinstance(comp, CoreComponent):
            raise IncorrectFolderUponLoad(
                f"Did not find a CoreComponent at {comp_folder_path} "
                f"when loading '{cct_name}'. Found "
                f"'{type(comp)}' instead.")
        if comp.get_type() != CoreComponentType[cct_name]:
            raise IncorrectFolderUponLoad(
                "Did not find the correct CoreComponent at "
                f"{comp_folder_path} for '{cct_name}'. Found "
                f"'{comp.get_type().name}' instead.")
        return comp

    def _init_components(self, model_load_path: Optional[str],
                         old_pipe: Optional['Pipeline']) -> None:
        (loaded_core_component_paths,
         loaded_addon_component_paths) = self._get_loaded_components_paths(
             model_load_path)
        for cct_name in self.config.components.comp_order:
            if cct_name in loaded_core_component_paths:
                comp = self._load_saved_core_component(
                    cct_name, loaded_core_component_paths.pop(cct_name))
            else:
                comp = self._init_component(
                    CoreComponentType[cct_name], model_load_path)
            self._components.append(comp)
        for addon_cnf in self.config.components.addons:
            addon = self._init_addon(
                addon_cnf, loaded_addon_component_paths, old_pipe)
            # mark as not dirty at loat / init time
            addon.config.mark_clean()
            self._addons.append(addon)

    def _get_loaded_addon_path(
            self, cnf: ComponentConfig,
            loaded_addon_component_paths: dict[tuple[str, str], str]
            ) -> Optional[str]:
        for key, folder in list(loaded_addon_component_paths.items()):
            comp_name, subname = key
            if comp_name != cnf.comp_name:
                continue
            if not isinstance(cnf, (ConfigMetaCAT, ConfigRelCAT)):
                raise UnkownAddonConfig(cnf, ConfigMetaCAT)
            if isinstance(cnf, ConfigMetaCAT):
                if cnf.general.category_name == subname:
                    del loaded_addon_component_paths[key]
                    return folder
            elif isinstance(cnf, ConfigRelCAT):
                if subname == 'rel_cat':
                    # NOTE: there can currently only ever be 1 RelCAT
                    #       this wasn't a limitation in v1
                    del loaded_addon_component_paths[key]
                    return folder
        return None

    def _load_addon(self, cnf: ComponentConfig, load_from: str
                    ) -> AddonComponent:
        # config is implicitly required argument
        model_load_path = os.path.dirname(os.path.dirname(load_from))
        addon = deserialise(
            load_from,
            # NOTE: the following are keyword arguments used
            #       for manual deserialisation
            cnf=cnf, tokenizer=self.tokenizer, cdb=self.cdb,
            vocab=self.vocab, model_load_path=model_load_path)
        if not isinstance(addon, AddonComponent):
            raise IncorrectAddonLoaded(
                f"Expected {AddonComponent.__name__}, but goet "
                f"{type(addon).__name__}")
        return addon

    def _init_addon(
            self, cnf: ComponentConfig,
            loaded_addon_component_paths: dict[tuple[str, str], str],
            old_pipe: Optional['Pipeline'],
            ) -> AddonComponent:
        if old_pipe:
            # If we are recreating a pipe and the addon configs haven't
            # changed then we can reuse existing addon instances.
            # This can be useful since otherwise we may need to load them
            # back off of disk, but we do not (necessarily) know where they
            # are or they may have changed after loading
            for old_addon in old_pipe._addons:
                if old_addon.config is cnf and not cnf.is_dirty:
                    return old_addon
                elif old_addon.config is cnf and cnf.is_dirty:
                    logger.warning(
                        "Not reusing existing addon '%s' because its config "
                        "has changed", old_addon.config.comp_name)
        loaded_path = self._get_loaded_addon_path(
            cnf, loaded_addon_component_paths)
        if loaded_path:
            return self._load_addon(cnf, loaded_path)
        return create_addon(
            cnf.comp_name, cnf=cnf, tokenizer=self.tokenizer, cdb=self.cdb,
            vocab=self.vocab, model_load_path=None)

    def get_doc(self, text: str) -> MutableDocument:
        """Get the document for this text.

        This essentially runs the tokenizer over the text.

        Args:
            text (str): The input text.

        Returns:
            MutableDocument: The resulting document.
        """
        doc = self._tokenizer(text)
        for comp in self._components:
            logger.info("Running component %s for %d of text (%s)",
                        comp.full_name, len(text), id(text))
            doc = comp(doc)
        for addon in self._addons:
            doc = addon(doc)
        return doc

    def entity_from_tokens(self, tokens: list[MutableToken]) -> MutableEntity:
        """Get the entity from the list of tokens.

        This effectively turns a list of (consecutive) documents
        into an entity.

        Args:
            tokens (list[MutableToken]): The tokens to use.

        Returns:
            MutableEntity: The resulting entity.
        """
        return self._tokenizer.entity_from_tokens(tokens)

    def get_component(self, ctype: CoreComponentType) -> CoreComponent:
        """Get the core component by the component type.

        Args:
            ctype (CoreComponentType): The core component type.

        Raises:
            ValueError: If no component by that type is found.

        Returns:
            CoreComponent: The corresponding core component.
        """
        for comp in self._components:
            if not comp.is_core() or not isinstance(comp, CoreComponent):
                continue
            if comp.get_type() is ctype:
                return comp
        raise ValueError(f"No component found of type {ctype}")

    def add_addon(self, addon: AddonComponent) -> None:
        self._addons.append(addon)
        # mark clean as of adding
        addon.config.mark_clean()

    def save_components(self,
                        serialiser_type: Union[AvailableSerialisers, str],
                        components_folder: str) -> None:
        for component in self.iter_all_components():
            if not isinstance(component, Serialisable):
                continue
            if not os.path.exists(components_folder):
                os.mkdir(components_folder)
            if isinstance(component, CoreComponent):
                comp_folder = os.path.join(
                    components_folder,
                    AbstractCoreComponent.NAME_PREFIX +
                    component.get_type().name)
            elif isinstance(component, AddonComponent):
                comp_folder = os.path.join(
                    components_folder,
                    f"{AddonComponent.NAME_PREFIX}{component.addon_type}"
                    f"{AddonComponent.NAME_SPLITTER}{component.name}")
            else:
                raise ValueError(
                    f"Unknown component: {type(component)} - does not appear "
                    "to be a CoreComponent or an AddonComponent")
            serialise(serialiser_type, component, comp_folder)

    def iter_all_components(self) -> Iterable[BaseComponent]:
        for component in self._components:
            yield component
        for addon in self._addons:
            yield addon

    def iter_addons(self) -> Iterable[AddonComponent]:
        yield from self._addons


class IncorrectArgumentsForTokenizer(TypeError):

    def __init__(self, provider: str):
        super().__init__(
            f"Incorrect arguments for tokenizer ({provider}).")


class IncorrectArgumentsForComponent(TypeError):

    def __init__(self, comp_type: CoreComponentType, comp_name: str):
        super().__init__(
            f"Incorrect arguments for core component {comp_type.name} "
            f"({comp_name}).")


class IncorrectCoreComponent(ValueError):

    def __init__(self, *args):
        super().__init__(*args)


class IncorrectFolderUponLoad(ValueError):

    def __init__(self, *args):
        super().__init__(*args)


class UnkownAddonConfig(ValueError):

    def __init__(self, cnf: ComponentConfig,
                 *existing_types: type[ComponentConfig]):
        super().__init__(
            f"Found unknown Addon config of type {type(cnf)}. "
            f"Existing types: {[etype.__name__ for etype in existing_types]}")


class IncorrectAddonLoaded(ValueError):

    def __init__(self, *args):
        super().__init__(*args)



import logging
import os
from typing import cast

from transformers import LlamaConfig

from medcat.config.config_rel_cat import ConfigRelCAT
from medcat.components.addons.relation_extraction.config import (
    RelExtrBaseConfig)

logger = logging.getLogger(__name__)


class RelExtrLlamaConfig(RelExtrBaseConfig):
    """ Class for LlamaConfig
    """

    name = 'llama-config'
    pretrained_model_name_or_path = "meta-llama/Llama-3.1-8B"
    hf_model_config: LlamaConfig

    @classmethod
    def load(cls, pretrained_model_name_or_path: str,
             relcat_config: ConfigRelCAT, **kwargs
             ) -> "RelExtrLlamaConfig":
        model_config = cls(pretrained_model_name_or_path, **kwargs)

        if pretrained_model_name_or_path and os.path.exists(
                pretrained_model_name_or_path):
            model_config.hf_model_config = cast(
                LlamaConfig, LlamaConfig.from_json_file(
                    pretrained_model_name_or_path))
            logger.info("Loaded config from file: %s",
                        pretrained_model_name_or_path)
        else:
            pretrained_model = relcat_config.general.model_name = (
                cls.pretrained_model_name_or_path)
            model_config.hf_model_config = cast(
                LlamaConfig, LlamaConfig.from_pretrained(
                    pretrained_model_name_or_path=pretrained_model, **kwargs))
            logger.info("Loaded config from pretrained: %s", pretrained_model)

        return model_config



from typing import Type, Generic, TypeVar, cast, Optional, Callable
import importlib
import logging


logger = logging.getLogger(__name__)

P = TypeVar('P')


class Registry(Generic[P]):
    def __init__(self, type: Type[P],
                 lazy_defaults: Optional[dict[str, tuple[str, str]]] = None
                 ) -> None:
        self._components: dict[str, Callable[..., P]] = {}
        self._type = type
        self._lazy_defaults = lazy_defaults.copy() if lazy_defaults else {}

    def register(self, component_name: str,
                 creator: Callable[..., P]):
        if component_name in self._components:
            prev = self._components[component_name]
            raise MedCATRegistryException(
                f"Component '{component_name}' already registered: {prev}")
        self._components[component_name] = creator

    def get_component(self, component_name: str
                      ) -> Callable[..., P]:
        """Get the component that's registered.

        The component generally refers to the class, but may
        be another method that creates the object needed.

        Args:
            component_name (str): The name of the component.

        Raises:
            MedCATRegistryException: If no component by requested name
                is registered.

        Returns:
            Callable[..., P]: The creator for the registered component.
        """
        # NOTE: some default implementations may be big imports,
        #       so we only want to import them if/when required.
        if component_name in self._lazy_defaults:
            self._ensure_lazy_default(component_name)
        if component_name not in self._components:
            raise MedCATRegistryException(
                f"No component registered by name '{component_name}'. "
                f"Available components: {self.list_components()}")
        return self._components[component_name]

    def _ensure_lazy_default(self, component_name: str) -> None:
        module_name, class_name = self._lazy_defaults.pop(component_name)
        logger.debug("Registering default %s '%s': '%s.%s'",
                     self._type.__name__, component_name, module_name,
                     class_name)
        module_in = importlib.import_module(module_name)
        if "." in class_name:
            cls_name, method_name = class_name.split(".")
        else:
            cls_name, method_name = class_name, None
        cls = getattr(module_in, cls_name)
        if method_name is not None:
            logger.debug("Using creator method %s.%s", cls_name, method_name)
            # use a creator method
            cls = getattr(cls, method_name)
        self.register(component_name, cast(Callable[..., P], cls))

    def register_all_defaults(self) -> None:
        """Register all default (lazily-added) components."""
        for comp_name in list(self._lazy_defaults):
            self._ensure_lazy_default(comp_name)

    def list_components(self) -> list[tuple[str, str]]:
        """List all available component names and class names.

        Returns:
            list[tuple[str, str]]: The list of the names and class names
                for each registered componetn.
        """
        comps = [(comp_name, comp.__name__)
                 for comp_name, comp in self._components.items()]
        for lazy_def_name, (_, lazy_def_class) in self._lazy_defaults.items():
            comps.append((lazy_def_name, lazy_def_class))
        return comps

    def unregister_component(self, component_name: str
                             ) -> Callable[..., P]:
        """Unregister a component.

        Args:
            component_name (str): The component name.

        Raises:
            MedCATRegistryException: If no component by the name specified
                had been registered.

        Returns:
            Callable[..., P]: The creator of the component.
        """
        if component_name not in self._components:
            raise MedCATRegistryException(
                f"No such component: {component_name}")
        logger.debug("Unregistering %s '%s'", self._type.__name__,
                     component_name)
        return self._components.pop(component_name)

    def unregister_all_components(self) -> None:
        """Unregister all components."""
        for comp_name in list(self._components):
            self.unregister_component(comp_name)

    def __contains__(self, component_name: str) -> bool:
        return (component_name in self._components or
                component_name in self._lazy_defaults)

    def __getitem__(self, component_name: str) -> Callable[..., P]:
        return self.get_component(component_name)


class MedCATRegistryException(Exception):

    def __init__(self, *args: object) -> None:
        super().__init__(*args)



from importlib.metadata import version as __version_method
from importlib.metadata import PackageNotFoundError as __PackageNotFoundError

try:
    __version__ = __version_method("medcat")
except __PackageNotFoundError:
    __version__ = "0.0.0-dev"



from typing import Callable, Iterable, TypeVar, Sized

T = TypeVar("T")


def _callback_iterator_iterable(identifier: str, data_iterator: Iterable[T],
                                callback: Callable[[str, int], None]
                                ) -> Iterable[T]:
    count = 0
    try:
        for item in data_iterator:
            count += 1
            yield item
    finally:
        callback(identifier, count)


def callback_iterator(identifier: str, data_iterator: Iterable[T],
                      callback: Callable[[str, int], None]) -> Iterable[T]:
    """Get an iterable with callback function to identify number of items.

    If the data has a size (i.e list or dict), the length of the data will
    be reported before iteration.

    If the data doesn't have a size (i.e a generator), the number of items
    iterated will be reported after the iteration is done.

    Args:
        identifier (str): The identifier / name for the iterator.
        data_iterator (Iterable[T]): The iterator.
        callback (Callable[[str, int], None]): The callback method.

    Returns:
        Iterable[T]: The wrapped iterator.
    """
    if isinstance(data_iterator, Sized):
        callback(identifier, len(data_iterator))
        return data_iterator
    return _callback_iterator_iterable(identifier, data_iterator, callback)



from typing import Iterable, Callable, Optional, Union, cast
import logging
from itertools import chain, repeat, islice
from tqdm import trange

from medcat.tokenizing.tokens import (MutableDocument, MutableEntity,
                                      MutableToken)
from medcat.cdb import CDB
from medcat.config.config import General, Preprocessing, CDBMaker
from medcat.utils.config_utils import temp_changed_config
from medcat.utils.data_utils import make_mc_train_test, get_false_positives
from medcat.utils.filters import project_filters
from medcat.data.mctexport import (
    MedCATTrainerExport, MedCATTrainerExportProject,
    MedCATTrainerExportDocument, count_all_annotations, iter_anns)
from medcat.preprocessors.cleaners import prepare_name, NameDescriptor
from medcat.components.types import CoreComponentType, TrainableComponent
from medcat.components.addons.addons import AddonComponent
from medcat.pipeline.pipeline import Pipeline


logger = logging.getLogger(__name__)


# NOTE: this should be used for changing the CDB, both for training and for
#       unlinking concept/names.
class Trainer:

    def __init__(self, cdb: CDB, caller: Callable[[str], MutableDocument],
                 pipeline: Pipeline):
        self.cdb = cdb
        self.config = cdb.config
        self.caller = caller
        self._pipeline = pipeline

    def train_unsupervised(self,
                           data_iterator: Iterable[str],
                           nepochs: int = 1,
                           fine_tune: bool = True,
                           progress_print: int = 1000,
                           #    checkpoint: Optional[Checkpoint] = None,
                           ) -> None:
        """Runs training on the data, note that the maximum length of a line
        or document is 1M characters. Anything longer will be trimmed.

        Args:
            data_iterator (Iterable):
                Simple iterator over sentences/documents, e.g. a open file
                or an array or anything that we can use in a for loop.
            nepochs (int):
                Number of epochs for which to run the training.
            fine_tune (bool):
                If False old training will be removed.
            progress_print (int):
                Print progress after N lines.
            checkpoint (Optional[medcat.utils.checkpoint.CheckpointUT]):
                The MedCAT checkpoint object
            is_resumed (bool):
                If True resume the previous training; If False, start a fresh
                new training.
        """
        with self.config.meta.prepare_and_report_training(
            data_iterator, nepochs, False
        ) as wrapped_iter:
            with temp_changed_config(self.config.components.linking,
                                     'train', True):
                self._train_unsupervised(wrapped_iter, nepochs, fine_tune,
                                         progress_print)

    def _train_unsupervised(self,
                            data_iterator: Iterable,
                            nepochs: int = 1,
                            fine_tune: bool = True,
                            progress_print: int = 1000,
                            #    checkpoint: Optional[Checkpoint] = None,
                            ) -> None:
        if not fine_tune:
            logger.info("Removing old training data!")
            self.cdb.reset_training()
        # checkpoint = self._init_ckpts(is_resumed, checkpoint)

        # latest_trained_step = checkpoint.count if checkpoint is
        # not None else 0
        latest_trained_step = 0  # TODO: add back checkpointing
        epochal_data_iterator = chain.from_iterable(repeat(data_iterator,
                                                           nepochs))
        for line in islice(epochal_data_iterator, latest_trained_step, None):
            if line is not None and line:
                # Convert to string
                line = str(line).strip()

                try:
                    _ = self.caller(line)
                except Exception as e:
                    logger.warning("LINE: '%s...' \t WAS SKIPPED", line[0:100])
                    logger.warning("BECAUSE OF:", exc_info=e)
            else:
                logger.warning("EMPTY LINE WAS DETECTED AND SKIPPED")

            latest_trained_step += 1
            if latest_trained_step % progress_print == 0:
                logger.info("DONE: %s", str(latest_trained_step))
            # if (checkpoint is not None and checkpoint.steps is not None
            #         and latest_trained_step % checkpoint.steps == 0):
            #     checkpoint.save(cdb=self.cdb, count=latest_trained_step)

    def _reset_cui_counts(self, train_set: MedCATTrainerExport,
                          reset_val: int = 100):
        # Get all CUIs
        cuis = []
        for project in train_set['projects']:
            for ann in (ann for doc in project['documents']
                        for ann in doc['annotations']):
                cuis.append(ann['cui'])
        for cui in set(cuis):
            if self.cdb.cui2info[cui]['count_train'] != 0:
                self.cdb.cui2info[cui]['count_train'] = reset_val

    def train_supervised_raw(self,
                             data: MedCATTrainerExport,
                             reset_cui_count: bool = False,
                             nepochs: int = 1,
                             print_stats: int = 0,
                             use_filters: bool = False,
                             terminate_last: bool = False,
                             use_overlaps: bool = False,
                             use_cui_doc_limit: bool = False,
                             test_size: float = 0,
                             devalue_others: bool = False,
                             use_groups: bool = False,
                             never_terminate: bool = False,
                             train_from_false_positives: bool = False,
                             extra_cui_filter: Optional[set[str]] = None,
                             #  checkpoint: Optional[Checkpoint] = None,
                             disable_progress: bool = False,
                             train_addons: bool = False,
                             ) -> tuple:
        """Train supervised based on the raw data provided.

        The raw data is expected in the following format:
        {'projects':
            [ # list of projects
                { # project 1
                    'name': '<some name>',
                    # list of documents
                    'documents': [{'name': '<some name>',  # document 1
                                    'text': '<text of the document>',
                                    # list of annotations
                                    'annotations': [# annotation 1
                                                    {'start': -1,
                                                    'end': 1,
                                                    'cui': 'cui',
                                                    'value': '<text value>'},
                                                    ...],
                                    }, ...]
                }, ...
            ]
        }

        Please take care that this is more a simulated online training then
        upervised.

        When filtering, the filters within the CAT model are used first,
        then the ones from MedCATtrainer (MCT) export filters,
        and finally the extra_cui_filter (if set).
        That is to say, the expectation is:
        extra_cui_filter ⊆ MCT filter ⊆ Model/config filter.

        Args:
            data (dict[str, list[dict[str, dict]]]):
                The raw data, e.g from MedCATtrainer on export.
            reset_cui_count (bool):
                Used for training with weight_decay (annealing). Each concept
                has a count that is there from the beginning of the CDB, that
                count is used for annealing. Resetting the count will
                significantly increase the training impact. This will reset
                the count only for concepts that exist in the the training
                data.
            nepochs (int):
                Number of epochs for which to run the training.
            print_stats (int):
                If > 0 it will print stats every print_stats epochs.
            use_filters (bool):
                Each project in medcattrainer can have filters, do we want to
                respect those filters
                when calculating metrics.
            terminate_last (bool):
                If true, concept termination will be done after all training.
            use_overlaps (bool):
                Allow overlapping entities, nearly always False as it is very
                difficult to annotate overlapping entities.
            use_cui_doc_limit (bool):
                If True the metrics for a CUI will be only calculated if that
                CUI appears in a document, in other words if the document was
                annotated for that CUI. Useful in very specific situations
                when during the annotation process the set of CUIs changed.
            test_size (float):
                If > 0 the data set will be split into train test based on
                this ration. Should be between 0 and 1. Usually 0.1 is fine.
            devalue_others(bool):
                Check add_name for more details.
            use_groups (bool):
                If True concepts that have groups will be combined and stats
                will be reported on groups.
            never_terminate (bool):
                If True no termination will be applied
            train_from_false_positives (bool):
                If True it will use false positive examples detected by medcat
                and train from them as negative examples.
            extra_cui_filter(Optional[set]):
                This filter will be intersected with all other filters, or if
                all others are not set then only this one will be used.
            checkpoint (Optional[Optional[medcat.utils.checkpoint.Checkpoint]):
                The MedCAT Checkpoint object
            disable_progress (bool):
                Whether to disable the progress output (tqdm). Defaults to
                False.
            train_addons (bool):
                Whether to also train the addons (e.g MetaCATs). Defaults
                to False.

        Returns:
            tuple: Consisting of the following parts
                fp (dict):
                    False positives for each CUI.
                fn (dict):
                    False negatives for each CUI.
                tp (dict):
                    True positives for each CUI.
                p (dict):
                    Precision for each CUI.
                r (dict):
                    Recall for each CUI.
                f1 (dict):
                    F1 for each CUI.
                cui_counts (dict):
                    Number of occurrence for each CUI.
                examples (dict):
                    FP/FN examples of sentences for each CUI.
        """
        # checkpoint = self._init_ckpts(is_resumed, checkpoint)

        # the config.linking.filters stuff is used directly in
# medcat.linking.context_based_linker and medcat.linking.vector_context_model
        # as such, they need to be kept up to date with per-project filters
        # However, the original state needs to be kept track of
        # so that it can be restored after training

        fp: dict[str, int] = {}
        fn: dict[str, int] = {}
        tp: dict[str, int] = {}
        p: dict[str, float] = {}
        r: dict[str, float] = {}
        f1: dict[str, float] = {}
        examples: dict[str, object] = {}

        cui_counts: dict[str, int] = {}

        if test_size == 0:
            logger.info("Running without a test set, or train==test")
            test_set = data
            train_set = data
        else:
            train_set, test_set, _, _ = make_mc_train_test(data, self.cdb,
                                                           test_size=test_size)

    # if print_stats > 0:
    #     fp, fn, tp, p, r, f1, cui_counts, examples = self._print_stats(
    #         test_set, use_project_filters=use_filters,
    #         use_cui_doc_limit=use_cui_doc_limit, use_overlaps=use_overlaps,
    #         use_groups=use_groups, extra_cui_filter=extra_cui_filter)
        if reset_cui_count:
            self._reset_cui_counts(train_set)

        # Remove entities that were terminated
        if not never_terminate:
            for ann in (ann for project in train_set['projects']
                        for doc in project['documents']
                        for ann in doc['annotations']):
                if ann.get('killed', False):
                    self.unlink_concept_name(ann['cui'], ann['value'], False)

        # latest_trained_step = (checkpoint.count if checkpoint is not None
        #                        else 0)
        # (current_epoch,
        #  current_project,
        #  current_document) = self._get_training_start(train_set,
        #                                               latest_trained_step)
        current_epoch = 0
        current_project = 0
        current_document = 0

        for epoch in trange(current_epoch, nepochs, initial=current_epoch,
                            total=nepochs, desc='Epoch', leave=False,
                            disable=disable_progress):
            self._perform_epoch(current_project, current_document, train_set,
                                disable_progress, extra_cui_filter,
                                use_filters, train_from_false_positives,
                                devalue_others, terminate_last,
                                never_terminate)

        # if print_stats > 0 and (epoch + 1) % print_stats == 0:
        #     fp, fn, tp, p, r, f1, cui_counts, examples = self._print_stats(
        #         test_set, epoch=epoch + 1, use_project_filters=use_filters,
        #         use_cui_doc_limit=use_cui_doc_limit,
        #         use_overlaps=use_overlaps, use_groups=use_groups,
        #         extra_cui_filter=extra_cui_filter)

        # # reset the state of filters
        # self.config.linking.filters = orig_filters

        if (train_addons and
                # NOTE if no annnotaitons, no point
                count_all_annotations(data) > 0):
            self._train_addons(data)

        return fp, fn, tp, p, r, f1, cui_counts, examples

    def _train_meta_cat(self, addon: AddonComponent,
                        data: MedCATTrainerExport) -> None:
        # NOTE: dynamic import to avoid circular imports
        from medcat.components.addons.meta_cat import MetaCATAddon
        _, _, ann0 = next(iter_anns(data))
        if not isinstance(addon, MetaCATAddon):
            raise TypeError(
                f"Expected MetaCATAddon, got {type(addon)}")
        if 'meta_anns' not in ann0:
            logger.info("No Meta Annotations found to train MetaCATs")
            return
        # only consider meta-cats that have been defined
        # for the category
        ann_names = ann0['meta_anns'].keys()  # type: ignore
        # adapt to alternative names if applicable
        cnf = addon.config
        cat_name = cnf.general.get_applicable_category_name(ann_names)
        if cat_name in ann_names:
            logger.debug("Training MetaCAT %s", cnf.general.category_name)
            # NOTE: this is a mypy quirk - the types are compatible
            addon.mc.train_raw(cast(dict, data))

    def _train_addons(self, data: MedCATTrainerExport):
        logger.info("Training addons within train_supervised_raw")
        for addon in self._pipeline._addons:
            if addon.addon_type == "meta_cat":
                self._train_meta_cat(addon, data)

    def _perform_epoch(self, current_project: int,
                       current_document: int,
                       train_set: MedCATTrainerExport,
                       disable_progress: bool,
                       extra_cui_filter: Optional[set[str]],
                       use_filters: bool,
                       train_from_false_positives: bool,
                       devalue_others: bool,
                       terminate_last: bool,
                       never_terminate: bool,
                       ) -> None:
        # Print acc before training
        for idx_project in trange(current_project,
                                  len(train_set['projects']),
                                  initial=current_project,
                                  total=len(train_set['projects']),
                                  desc='Project', leave=False,
                                  disable=disable_progress):
            project = train_set['projects'][idx_project]
            with project_filters(
                    self.config.components.linking.filters, project,
                    extra_cui_filter, use_filters):
                self._train_supervised_for_project(
                    project, current_document, train_from_false_positives,
                    devalue_others)

        if terminate_last and not never_terminate:
            # Remove entities that were terminated,
            # but after all training is done
            # for project in train_set['projects']:
            #     for doc in project['documents']:
            #         for ann in doc_annotations:
            for ann in (ann for project in train_set['projects']
                        for doc in project['documents']
                        for ann in doc['annotations']):
                if ann.get('killed', False):
                    self.unlink_concept_name(ann['cui'], ann['value'], False)

    def _train_supervised_for_project(self,
                                      project: MedCATTrainerExportProject,
                                      current_document: int,
                                      train_from_false_positives: bool,
                                      devalue_others: bool):
        with self.config.meta.prepare_and_report_training(
                project['documents'], 1, True, project_name=project['name']
                ) as docs:
            with temp_changed_config(self.config.components.linking,
                                     'train', True):
                self._train_supervised_for_project2(
                    docs, current_document, train_from_false_positives,
                    devalue_others)

    def _train_supervised_for_project2(self,
                                       docs: list[MedCATTrainerExportDocument],
                                       current_document: int,
                                       train_from_false_positives: bool,
                                       devalue_others: bool):
        cnf_linking = self.config.components.linking
        for idx_doc in trange(current_document,
                              len(docs),
                              initial=current_document,
                              total=len(docs),
                              desc='Document', leave=False):
            doc = docs[idx_doc]
            with temp_changed_config(self.config.components.linking,
                                     'train', False):
                mut_doc = self.caller(doc['text'])

            # Compatibility with old output where annotations are a list
            for ann in doc['annotations']:
                if ann.get('killed', False):
                    continue
                logger.info("    Annotation %s (%s) [%d:%d]",
                            ann['value'], ann['cui'], ann['start'], ann['end'])
                cui = ann['cui']
                start = ann['start']
                end = ann['end']
                mut_entity = mut_doc.get_tokens(start, end)
                if not mut_entity:
                    logger.warning(
                        "When looking for CUI '%s' (value '%s') [%d...%d] "
                        "within the document '%s' (ID %s) was unable "
                        "to get any tokens that match the start and end. ",
                        cui, ann['value'], start, end,
                        doc['name'], doc['id'])
                    continue
                deleted = bool(ann.get('deleted', False))
                if not cnf_linking.filters.check_filters(cui):
                    continue
                try:
                    self.add_and_train_concept(
                        cui=cui, name=ann['value'], mut_doc=mut_doc,
                        mut_entity=mut_entity, negative=deleted,
                        devalue_others=devalue_others)
                except (ValueError, KeyError) as ve:
                    context_window = 20  # characters
                    splitter_left, splitter_right = "<", ">"
                    cur_text = doc['text']
                    context_start = max(start - context_window, 0)
                    context_end = min(end + context_window, len(cur_text) - 1)
                    context = (cur_text[context_start: start] +
                               splitter_left +
                               cur_text[start: end] +
                               splitter_right +
                               cur_text[end: context_end])
                    if context_start > 0:
                        context = "[...]" + context
                    if context_end < len(cur_text) - 1:
                        context += "[...]"
                    raise ValueError(
                        f"Failed to identify '{cui}' ({ann['value']}) "
                        f"([{ann['start']}:{ann['end']}]) "
                        f"in '{context}' {mut_entity} within document "
                        f"{doc['id']} | {doc['name']}") from ve
            if train_from_false_positives:
                fps: list[MutableEntity] = get_false_positives(doc, mut_doc)

                for fp in fps:  # type: ignore
                    fp_: MutableEntity = fp  # type: ignore
                    # TODO: allow adding/training
                    self.add_and_train_concept(
                        cui=fp_.cui, name=fp_.base.text,
                        mut_doc=mut_doc, mut_entity=fp_,
                        negative=True, do_add_concept=False)

            # latest_trained_step += 1
            # if (checkpoint is not None and checkpoint.steps is not None
            #         and latest_trained_step % checkpoint.steps == 0):
            #     checkpoint.save(self.cdb, latest_trained_step)

    def unlink_concept_name(self, cui: str, name: str,
                            preprocessed_name: bool = False) -> None:
        """Unlink a concept name from the CUI (or all CUIs if full_unlink),
        removes the link from the Concept Database (CDB). As a consequence
        medcat will never again link the `name` to this CUI - meaning the
        name will not be detected as a concept in the future.

        Args:
            cui (str):
                The CUI from which the `name` will be removed.
            name (str):
                The span of text to be removed from the linking dictionary.
            preprocessed_name (bool):
                Whether the name being used is preprocessed.

        Examples:

            >>> # To never again link C0020538 to HTN
            >>> cat.unlink_concept_name('C0020538', 'htn', False)
        """

        cuis = [cui]
        if preprocessed_name:
            names: dict[str, NameDescriptor] = {
                name: NameDescriptor([], set(), name, name.isupper())}
        else:
            names = prepare_name(name, self._pipeline.tokenizer, {},
                                 self._pn_configs)

        # If full unlink find all CUIs
        if self.config.general.full_unlink:
            logger.warning("In the config `full_unlink` is set to `True`. "
                           "Thus removing all CUIs linked to the specified "
                           "name (%s)", name)
            for n in names:
                if n not in self.cdb.name2info:
                    continue
                cuis.extend(self.cdb.name2info[n]['per_cui_status'].keys())

        # Remove name from all CUIs
        for c in cuis:
            self.cdb._remove_names(cui=c, names=names.keys())

    def add_and_train_concept(self,
                              cui: str,
                              name: str,
                              mut_doc: Optional[MutableDocument] = None,
                              mut_entity: Optional[
                                  Union[list[MutableToken],
                                        MutableEntity]] = None,
                              ontologies: set[str] = set(),
                              name_status: str = 'A',
                              type_ids: set[str] = set(),
                              description: str = '',
                              full_build: bool = True,
                              negative: bool = False,
                              devalue_others: bool = False,
                              do_add_concept: bool = True) -> None:
        r"""Add a name to an existing concept, or add a new concept, or do not
        do anything if the name or concept already exists. Perform training if
        spacy_entity and spacy_doc are set.

        Args:
            cui (str):
                CUI of the concept.
            name (str):
                Name to be linked to the concept (in the case of MedCATtrainer
                this is simply the selected value in text, no preprocessing or
                anything needed).
            mut_doc (Optional[MutableDocument]):
                Spacy representation of the document that was manually
                annotated.
            mut_entity (mut_entity: Optional[Union[list[MutableToken],
                                                   MutableEntity]]):
                Given the spacy document, this is the annotated span of text -
                list of annotated tokens that are marked with this CUI.
            ontologies (set[str]):
                ontologies in which the concept exists (e.g. SNOMEDCT, HPO)
            name_status (str):
                One of `P`, `N`, `A`
            type_ids (set[str]):
                Semantic type identifier (have a look at TUIs in UMLS or
                SNOMED-CT)
            description (str):
                Description of this concept.
            full_build (bool):
                If True the dictionary self.addl_info will also be populated,
                contains a lot of extra information about concepts, but can be
                very memory consuming. This is not necessary for normal
                functioning of MedCAT (Default Value `False`).
            negative (bool):
                Is this a negative or positive example.
            devalue_others (bool):
                If set, cuis to which this name is assigned and are not `cui`
                will receive negative training given that negative=False.
            do_add_concept (bool):
                Whether to add concept to CDB.
        """
        names = prepare_name(name, self._pipeline.tokenizer_with_tag, {},
                             self._pn_configs)
        if (not names and cui not in self.cdb.cui2info and
                name_status == 'P'):
            logger.warning(
                "No names were able to be prepared in "
                "CAT.add_and_train_concept method. As such no preferred name "
                "will be able to be specifeid. The CUI: '%s' and raw name: "
                "'%s'", cui, name)
        # Only if not negative, otherwise do not add the new name if in fact
        # it should not be detected
        if do_add_concept and not negative:
            self.cdb._add_concept(cui=cui, names=names, ontologies=ontologies,
                                  name_status=name_status, type_ids=type_ids,
                                  description=description,
                                  full_build=full_build)

        if mut_entity is None or mut_doc is None:
            return
        linker = self._pipeline.get_component(
            CoreComponentType.linking)
        if not isinstance(linker, TrainableComponent):
            logger.warning(
                "Linker cannot be trained during add_and_train_concept"
                "because it has no train method: %s", linker)
        else:
            # Train Linking
            if isinstance(mut_entity, list):
                mut_entity = self._pipeline.entity_from_tokens(mut_entity)
            linker.train(cui=cui, entity=mut_entity, doc=mut_doc,
                         negative=negative, names=names)

            if not negative and devalue_others:
                # Find all cuis
                cuis: set[str] = set()
                for n in names:
                    if n in self.cdb.name2info:
                        info = self.cdb.name2info[n]
                        cuis.update(info['per_cui_status'].keys())
                # Remove the cui for which we just added positive training
                if cui in cuis:
                    cuis.remove(cui)
                # Add negative training for all other CUIs that link to
                # these names
                for _cui in cuis:
                    linker.train(cui=_cui, entity=mut_entity, doc=mut_doc,
                                 negative=True)

    @property
    def _pn_configs(self) -> tuple[General, Preprocessing, CDBMaker]:
        return (self.config.general, self.config.preprocessing,
                self.config.cdb_maker)



from typing import Union, overload
import numpy as np
from numpy.typing import NDArray


# NOTE: this would be faster with `gensim`
#       since it uses BLAS, but can't get it to
#       work well with numpy2/scipy combo
def unitvec(vec: np.ndarray) -> np.ndarray:
    """Get the unitvector.

    Args:
        vec (np.ndarray): The non-unit vector.

    Returns:
        np.ndarray: The new unit vector.
    """
    return vec / np.linalg.norm(vec)


@overload
def sigmoid(x: float) -> float:
    pass


@overload
def sigmoid(x: NDArray[np.float64]) -> NDArray[np.float64]:
    pass


def sigmoid(x: Union[np.ndarray, float]) -> Union[np.ndarray, float]:
    return 1 / (1 + np.exp(-x))



from typing import Optional
import os
import json
import logging

import torch

from medcat.components.addons.meta_cat import MetaCAT, MetaCATAddon
from medcat.components.addons.meta_cat.mctokenizers.tokenizers import (
    TokenizerWrapperBase, load_tokenizer)
from medcat.config.config_meta_cat import ConfigMetaCAT
from medcat.tokenizing.tokenizers import BaseTokenizer

from medcat.utils.legacy.helpers import fix_old_style_cnf


logger = logging.getLogger(__name__)


def _load_legacy(config: ConfigMetaCAT, save_dir_path: str) -> MetaCAT:
    tokenizer: Optional[TokenizerWrapperBase] = None
    # Load tokenizer
    tokenizer = load_tokenizer(config, save_dir_path)

    # Create meta_cat
    meta_cat = MetaCAT(tokenizer=tokenizer, embeddings=None, config=config)

    # Load the model
    model_save_path = os.path.join(save_dir_path, 'model.dat')
    device = torch.device(config.general.device)
    if not torch.cuda.is_available() and device.type == 'cuda':
        logger.warning(
            'Loading a MetaCAT model without GPU availability, '
            'stored config used GPU')
        config.general.device = 'cpu'
        device = torch.device('cpu')
    meta_cat.model.load_state_dict(
        torch.load(model_save_path, map_location=device))
    return meta_cat


def load_cnf(cnf_path: str) -> ConfigMetaCAT:
    with open(cnf_path) as f1:
        data = json.load(f1)
    data = fix_old_style_cnf(data)
    cnf = ConfigMetaCAT.model_validate(data)
    cnf.comp_name = MetaCATAddon.addon_type
    return cnf


def get_meta_cat_from_old(old_path: str, tokenizer: BaseTokenizer
                          ) -> MetaCATAddon:
    """Convert a v1 MetaCAT folder to a v2 MetaCAT.

    Args:
        old_path (str): The v1 MetaCAT file path.
        tokenizer (BaseTokenizer): The tokenizer.

    Returns:
        MetaCATAddon: The v2 MetaCAT.
    """
    cnf = load_cnf(os.path.join(old_path, "config.json"))
    mc = _load_legacy(cnf, old_path)
    addon = MetaCATAddon.create_new(cnf, tokenizer)
    addon._mc = mc
    return addon




from typing import List, Union
import pandas as pd
import tqdm
import os
from typing import Dict

_DEFAULT_COLUMNS: list = [
    "CUI",
    "LAT",
    "TS",
    "LUI",
    "STT",
    "SUI",
    "ISPREF",
    "AUI",
    "SAUI",
    "SCUI",
    "SDUI",
    "SAB",
    "TTY",
    "CODE",
    "STR",
    "SRL",
    "SUPPRESS",
    "CVF",
]

_DEFAULT_SEM_TYPE_COLUMNS: list = [
    "CUI",
    "TUI",
    "STN",
    "STY",
    "ATUI",
    "CVF",
]

_DEFAULT_MRHIER_COLUMNS: list = [
    "CUI",
    "AUI",
    "CXN",
    "PAUI",
    "SAB",
    "RELA",
    "PTR",
    "HCD",
    "CVF",
]

medcat_csv_mapper: dict = {
    'CUI': 'cui',
    'STR': 'name',
    'SAB': 'ontologies',
    'ISPREF': 'name_status',
    'TUI': 'type_ids',  # from MRSTY.RRF
}


class UMLS:
    """Pre-process UMLS release files:
    Args:
        main_file_name (str):
            Path to the main file name (probably MRCONSO.RRF)
        sem_types_file (str):
            Path to the semantic types file name (probably MRSTY.RRF)
        allow_langugages (list):
            Languages to filter out. Defaults to just English (['ENG']).
        sep (str):
            The separator used within the files. Defaults to '|'.
    """

    def __init__(self, main_file_name: str, sem_types_file: str,
                 allow_languages: list = ['ENG'], sep: str = '|'):
        self.main_file_name = main_file_name
        self.sem_types_file = sem_types_file
        self.main_columns = list(_DEFAULT_COLUMNS)  # copy
        self.sem_types_columns = list(_DEFAULT_SEM_TYPE_COLUMNS)  # copy
        self.mrhier_columns = list(_DEFAULT_MRHIER_COLUMNS)  # copy
        self.sep = sep
        # copy in case of default list
        self.allow_langugages = list(
            allow_languages) if allow_languages else allow_languages

    def to_concept_df(self) -> pd.DataFrame:
        """Create a concept DataFrame.
        The default column names are expected.

        Returns:
            pd.DataFrame: The resulting DataFrame
        """
        # target columns:
        # cui, name, name_status, ontologies, description_type_ids, type_ids
        df = pd.read_csv(self.main_file_name,
                         names=self.main_columns, sep=self.sep,
                         index_col=False)

        # filter languages
        if self.allow_langugages:
            df = df[df["LAT"].isin(self.allow_langugages)]

        # TODO filter by activity ?

        # get TUI

        sem_types = pd.read_csv(
            self.sem_types_file, names=self.sem_types_columns, sep=self.sep,
            index_col=False)
        df = df.merge(sem_types)

        # rename columns

        df = df.rename(columns=medcat_csv_mapper)

        # pop all unnecessary columns

        # all initial columns should have been renamed
        for col_name in self.main_columns + self.sem_types_columns:
            if col_name in df.columns:
                df.pop(col_name)

        # looks like description_type_ids is not really used anywhere,
        # so I won't look for it

        return df

    def map_umls2snomed(self) -> pd.DataFrame:
        """Map to SNOMED-CT.

        Currently, uses the SCUI column. At the time of writing, this is equal
        to the CODE column.
        But this may not be the case in the future.

        Returns:
            pd.DataFrame: Dataframe that contains the SCUI (source CUI) as
                well as the UMLS CUI for each applicable concept
        """
        df = pd.read_csv(self.main_file_name, names=self.main_columns,
                         sep=self.sep, index_col=False, dtype={'SCUI': 'str'})
        # get only SNOMED-CT US based concepts that have a SNOMED-CT (source)
        # CUI
        df = df[df.SAB == 'SNOMEDCT_US'][df.SCUI.notna()]
        # sort by SCUI
        df = df.sort_values(by='SCUI').reset_index(drop=True)
        # rearrange with SCUI as the first column
        df = df[['SCUI',] + [
            col for col in df.columns.values if col != 'SCUI']]
        return df

    def map_umls2icd10(self) -> pd.DataFrame:
        """Map to ICD-10.

        Available SAB's that contain 'ICD10':
         - CCSR_ICD10CM - CCSR_ICD10CM (Clinical Classifications Software Refined for ICD-10-CM) - Synopsis
         - CCSR_ICD10PCS - CCSR_ICD10PCS (Clinical Classifications Software Refined for ICD-10-PCS) - Synopsis
         - DMDICD10 - DMDICD10 (ICD-10 German) - Statistics
         - ICD10AE - ICD10AE (ICD-10, American English Equivalents) - Synopsis
         - ICD10AMAE - ICD10AMAE (ICD-10, Australian Modification, Americanized English Equivalents) - Synopsis
         - ICD10AM - ICD10AM (ICD-10, Australian Modification) - Synopsis
         - ICD10DUT - ICD10DUT (ICD10, Dutch Translation) - Synopsis
         - ICD10PCS - ICD10PCS (ICD-10 Procedure Coding System) - Synopsis
         - ICD10 - ICD10 (International Classification of Diseases and Related Health Problems, Tenth Revision) - Synopsis
         - ICPC2ICD10DUT - ICPC2ICD10DUT (ICPC2-ICD10 Thesaurus, Dutch Translation) - Synopsis
         - ICPC2ICD10ENG - ICPC2ICD10ENG (ICPC2-ICD10 Thesaurus) - Synopsis
         - MTHICPC2ICD10AE - MTHICPC2ICD10AE (ICPC2E-ICD10 Thesaurus, American English Equivalents) - Synopsis

        Currently only using 'ICD10'. But others may be relevant as well.

        If one wants to use one of the other sources listed above,
        they would need to use the map_umls2source method.

        Returns:
            pd.DataFrame: DataFrame that has the ICD-10 codes
        """  # noqa
        return self.map_umls2source(sources='ICD10')

    def map_umls2source(self, sources: Union[str, List[str]]) -> pd.DataFrame:
        """Allows mapping to an arbitrary

        Args:
            sources (Union[str, List[str]]): The source or sources to include.

        Returns:
            pd.DataFrame: DataFrame that has the target source codes
        """
        df = pd.read_csv(self.main_file_name, names=self.main_columns,
                         sep=self.sep, index_col=False, dtype={'CODE': 'str'})
        # get the specified source(s)
        if isinstance(sources, list):
            df = df[df.SAB.isin(sources)][df.CODE.notna()]
        else:
            df = df[df.SAB == sources][df.CODE.notna()]
        # sort by CODE
        df = df.sort_values(by='CODE').reset_index(drop=True)
        # rearrange columns starting with CODE
        df = df[['CODE',] + [
            col for col in df.columns.values if col != 'CODE']]
        return df

    def get_pt2ch(self) -> dict:
        """Generates a parent to children dict.

        It goes through all the < # TODO

        The resulting dictionary maps a CUI to a list of CUIs that
        consider that CUI as their parent.

        PS:
        This expects the MRHIER.RRF file to also exist in the same folder
        as the MRCONSO.RRF file.

        Raises:
            ValueError: If the MRHIER.RRF file wasn't found

        Returns:
            dict: The dictionary of parent CUI and their children.
        """
        path = self.main_file_name.rsplit('/', 1)[0]
        hier_file = f"{path}/MRHIER.RRF"

        if not os.path.exists(hier_file):
            raise ValueError(
                'Expected MRHIER.RRF to exist within the same parent folder '
                f'({path})')

        conso_df = pd.read_csv(self.main_file_name, names=self.main_columns,
                               sep=self.sep, index_col=False)

        hier_df = pd.read_csv(hier_file, sep=self.sep, index_col=False,
                              header=None, names=self.mrhier_columns)

        # filter languages
        if self.allow_langugages:
            conso_df = conso_df[conso_df["LAT"].isin(self.allow_langugages)]

        # create a AUI -> CUI map
        aui_cui = dict(zip(conso_df["AUI"], conso_df["CUI"]))

        # remove non-preferred from conso
        conso_df = conso_df[conso_df['ISPREF'] == 'Y']

        # filter ISA relationships
        hier_df = hier_df[hier_df['RELA'] == 'isa']

        # merge dataframes
        merged_df = pd.merge(conso_df, hier_df, on=['AUI', 'CUI'])

        # only keep CUI and parent AUI
        cui_parent = merged_df[['CUI', 'PAUI']]
        # only include CUIs with a parent
        cui_parent = cui_parent[cui_parent['PAUI'].notna()]

        # create dict
        pt2ch: dict = {}
        for _, row in tqdm.tqdm(cui_parent.iterrows(),
                                total=len(cui_parent.index)):
            cur_cui = row['CUI']
            paui = row['PAUI']
            parent_cui = aui_cui[paui]
            # avoid self as parent/child
            if parent_cui == cur_cui:
                continue
            if parent_cui not in pt2ch:
                pt2ch[parent_cui] = set()
            pt2ch[parent_cui].add(cur_cui)
        # move from set to list for consistency with SNOMED
        pt2ch: Dict[str, List[str]] = pt2ch  # type: ignore
        for k, v in pt2ch.items():
            pt2ch[k] = list(v)
        return pt2ch


if __name__ == '__main__':
    import sys
    if len(sys.argv) < 3:
        print('Need to specify two file locations: MRCONSO.RRF and MRSTY.RRF')
        sys.exit(1)
    umls = UMLS(sys.argv[1], sys.argv[2])
    df = umls.to_concept_df()
    save_file = "preprocessed_umls.csv"
    print(f"Saving to {save_file}")
    df.to_csv(save_file, index=False)
    print('Converting to SNOMED')
    to_snomed = umls.map_umls2snomed()
    print('As SNOMED:')
    print(to_snomed.head())
    to_ICD10 = umls.map_umls2icd10()
    print('As ICD-10:')
    print(to_ICD10.head())
    to_ICD10_man = umls.map_umls2source(sources=['ICD10'])
    print('As ICD-10(MAN):')
    print(to_ICD10_man.head())
    pt2ch = umls.get_pt2ch()
    print('Get parent-child dict', len(pt2ch),
          '' if len(pt2ch) > 1_000 else pt2ch)
    all_vals = [len(v) for v in pt2ch.values()]
    print('LEN of VALS:', sum(all_vals), 'max',
          max(all_vals), 'min', min(all_vals), 'mean',
          sum(all_vals) / len(all_vals))
    import random
    random_4_keys = random.sample(list(pt2ch.keys()), k=4)

    def _get_name(cui: str) -> str:
        matches = df[df['cui'] == cui]
        if len(matches.index) == 0:
            return 'N/A'  # UNKNOWN
        return matches['name'].iloc[0]
    print('FF RAW   ', [f"{k}:{pt2ch[k]}" for k in random_4_keys])
    print('FIRST FEW', [
        (f"{_get_name(key)} ({key})", [f"{_get_name(child)} ({child})"
                                       for child in pt2ch[key]])
        for key in random_4_keys])



import os


def ensure_folder_if_parent(folder_name: str) -> None:
    """Ensure the folder exists if its parent folder exists.

    Create a folder if the parent folder exists.
    If the parent folder does not exist, raise an error.

    Args:
        folder_name (str): The target folder.

    Raises:
        ValueError: If the parent folder does not exist.
    """
    target_folder = os.path.dirname(folder_name)
    if (os.path.exists(target_folder) and
            not os.path.exists(folder_name)):
        os.makedirs(folder_name)
    elif not os.path.exists(target_folder):
        raise ValueError("The target folder does not exist: "
                         f"{target_folder}")



from typing import Optional, Callable, cast

from tqdm import tqdm
import traceback

from medcat.cat import CAT
from medcat.utils.filters import project_filters
from medcat.data.mctexport import (
    MedCATTrainerExport, MedCATTrainerExportProject,
    MedCATTrainerExportDocument, MedCATTrainerExportAnnotation)
from medcat.config.config import LinkingFilters
from medcat.cdb.concepts import CUIInfo, get_new_cui_info
from medcat.tokenizing.tokens import MutableEntity, MutableDocument


class StatsBuilder:

    def __init__(self,
                 filters: LinkingFilters,
                 addl_info: dict,
                 doc_getter: Callable[[str], Optional[MutableDocument]],
                 cui2info: dict[str, CUIInfo],
                 use_project_filters: bool = False,
                 use_overlaps: bool = False,
                 #  use_cui_doc_limit: bool = False,
                 #  use_groups: bool = False,
                 extra_cui_filter: Optional[set[str]] = None) -> None:
        self.filters = filters
        self.addl_info = addl_info
        self.doc_getter = doc_getter
        self.cui2info = cui2info
        self.use_project_filters = use_project_filters
        self.use_overlaps = use_overlaps
        # self.use_cui_doc_limit = use_cui_doc_limit
        # self.use_groups = use_groups
        self.extra_cui_filter = extra_cui_filter
        self._reset_stats()

    def _reset_stats(self):
        self.tp = 0
        self.fp = 0
        self.fn = 0
        self.fps: dict[str, int] = {}
        self.fns: dict[str, int] = {}
        self.tps: dict[str, int] = {}
        self.cui_prec: dict[str, float] = {}
        self.cui_rec: dict[str, float] = {}
        self.cui_f1: dict[str, float] = {}
        self.cui_counts: dict[str, int] = {}
        self.examples: dict = {'fp': {}, 'fn': {}, 'tp': {}}
        self.fp_docs: set = set()
        self.fn_docs: set = set()

    def process_project(self, project: MedCATTrainerExportProject) -> None:
        """Process the project.

        This processes each document in the project.

        Args:
            project (MedCATTrainerExportProject): The trainer export project.
        """
        project_name = cast(str, project.get('name'))
        project_id = cast(str, project.get('id'))

        documents = project["documents"]
        for dind, doc in tqdm(
            enumerate(documents),
            desc="Stats document",
            total=len(documents),
            leave=False,
        ):
            self.process_document(project_name, project_id, doc)

    def process_document(self, project_name: str, project_id: str,
                         doc: MedCATTrainerExportDocument
                         ) -> None:
        """Process the trainer export document.

        Args:
            project_name (str): The project within which this document lies.
            project_id (str): The project ID for the project.
            doc (MedCATTrainerExportDocument): The trainer export document.
        """
        anns = doc['annotations']

        # Apply document level filtering, in this case project_filter is
        # ignored while the extra_cui_filter is respected still
        # if self.use_cui_doc_limit:
        #     _cuis = set([ann['cui'] for ann in anns])
        #     if _cuis:
        #         self.filters.cuis = intersect_nonempty_set(_cuis,
        #             self.extra_cui_filter)
        #     else:
        #         self.filters.cuis = {'empty'}

        mut_doc: MutableDocument = self.doc_getter(
            doc['text'])  # type: ignore

        p_anns = mut_doc.linked_ents  # or all ents?

        (anns_norm, anns_norm_neg,
         anns_examples, _) = self._preprocess_annotations(
             project_name, project_id, doc, anns)

        p_anns_norm, p_anns_examples = self._process_p_anns(
            project_name, project_id, doc, p_anns)
        self._count_p_anns_norm(doc, anns_norm, anns_norm_neg,
                                p_anns_norm, p_anns_examples)
        self._process_anns_norm(doc, anns_norm, p_anns_norm, anns_examples)

    def _process_anns_norm(self, doc: MedCATTrainerExportDocument,
                           anns_norm: list[tuple[int, str]],
                           p_anns_norm: list[tuple[int, str]],
                           anns_examples: list[dict]) -> None:
        for iann, ann in enumerate(anns_norm):
            if ann not in p_anns_norm:
                cui = ann[1]
                self.fn += 1
                self.fn_docs.add(doc.get('name', 'unk'))

                self.fns[cui] = self.fns.get(cui, 0) + 1
                examples = self.examples['fn'].get(cui, [])
                self.examples['fn'][cui] = examples + [anns_examples[iann]]

    def _process_p_anns(self, project_name: str, project_id: str,
                        doc: MedCATTrainerExportDocument,
                        p_anns: list[MutableEntity]
                        ) -> tuple[list[tuple[int, str]], list[dict]]:
        p_anns_norm: list[tuple[int, str]] = []
        p_anns_examples: list[dict] = []
        for ann in p_anns:
            cui = ann.cui

            p_anns_norm.append((ann.base.start_char_index, cui))
            p_anns_examples.append(self._create_annotation_2(
                project_name, project_id, cui, doc, ann))
        return p_anns_norm, p_anns_examples

    def _count_p_anns_norm(self, doc: MedCATTrainerExportDocument,
                           anns_norm: list[tuple[int, str]],
                           anns_norm_neg: list[tuple[int, str]],
                           p_anns_norm: list[tuple[int, str]],
                           p_anns_examples: list[dict]) -> None:
        for iann, ann in enumerate(p_anns_norm):
            cui = ann[1]
            if ann in anns_norm:
                self.tp += 1
                self.tps[cui] = self.tps.get(cui, 0) + 1

                example = p_anns_examples[iann]

                examples = self.examples['tp'].get(cui, [])
                self.examples['tp'][cui] = examples + [example]
            else:
                self.fp += 1
                self.fps[cui] = self.fps.get(cui, 0) + 1
                self.fp_docs.add(doc.get('name', 'unk'))

                # Add example for this FP prediction
                example = p_anns_examples[iann]
                if ann in anns_norm_neg:
                    # Means that it really was annotated as negative
                    example['real_fp'] = True

                examples = self.examples['fp'].get(cui, [])
                self.examples['fp'][cui] = examples + [example]

    def _create_annotation(self, project_name: str, project_id: str, cui: str,
                           doc: MedCATTrainerExportDocument,
                           ann: MedCATTrainerExportAnnotation) -> dict:
        return {"text": doc['text'][max(0, ann['start'] - 60):ann['end'] + 60],
                "cui": cui,
                "start": ann['start'],
                "end": ann['end'],
                "source value": ann['value'],
                "acc": 1,
                "project name": project_name,
                "document name": doc.get('name'),
                "project id": project_id,
                "document id": doc.get('id')}

    def _create_annotation_2(self, project_name: str, project_id: str,
                             cui: str, doc: MedCATTrainerExportDocument,
                             ann: MutableEntity) -> dict:
        start = max(0, ann.base.start_char_index - 60)
        end = ann.base.end_char_index + 60
        return {"text": doc['text'][start:end],
                "cui": cui,
                "start": ann.base.start_char_index,
                "end": ann.base.start_char_index,
                "source value": ann.base.text,
                "acc": float(ann.context_similarity),
                "project name": project_name,
                "document name": doc.get('name'),
                "project id": project_id,
                "document id": doc.get('id')}

    def _preprocess_annotations(self, project_name: str, project_id: str,
                                doc: MedCATTrainerExportDocument,
                                anns: list[MedCATTrainerExportAnnotation]
                                ) -> tuple[list[tuple[int, str]],
                                           list[tuple[int, str]],
                                           list[dict],
                                           list[str]]:
        anns_norm: list[tuple[int, str]] = []
        anns_norm_neg: list[tuple[int, str]] = []
        anns_examples: list[dict] = []
        anns_norm_cui: list[str] = []
        for ann in anns:
            cui = ann['cui']
            if self.filters.check_filters(cui):

                if (ann.get('validated', True) and
                    (not ann.get('killed', False) and not ann.get('deleted',
                                                                  False))):
                    anns_norm.append((ann['start'], cui))
                    anns_examples.append(self._create_annotation(
                        project_name, project_id, cui, doc, ann))
                elif (ann.get('validated', True) and
                      (ann.get('killed', False) or ann.get('deleted', False))):
                    anns_norm_neg.append((ann['start'], cui))

                if ann.get("validated", True):
                    # This is used to test was someone annotating for this
                    # CUI in this document
                    anns_norm_cui.append(cui)
                    self.cui_counts[cui] = self.cui_counts.get(cui, 0) + 1
        return anns_norm, anns_norm_neg, anns_examples, anns_norm_cui

    def finalise_report(self, epoch: int, do_print: bool = True):
        """Finalise the report / metrics.

        This prints out the overall metrics and calculates per CUI metrics.

        Args:
            epoch (int): The number of the current epoch.
            do_print (bool, optional): Whether to print the output.
                Defaults to True.
        """
        try:
            prec = self.tp / (self.tp + self.fp)
            rec = self.tp / (self.tp + self.fn)
            f1 = 2 * (prec * rec) / (prec + rec)
            if do_print:
                print("Epoch: {}, Prec: {}, Rec: {}, F1: {}\n".format(
                    epoch, prec, rec, f1))
                print("Docs with false positives: {}\n".format("; ".join(
                    [str(x) for x in list(self.fp_docs)[0:10]])))
                print("Docs with false negatives: {}\n".format("; ".join(
                    [str(x) for x in list(self.fn_docs)[0:10]])))

            # Sort fns & prec
            fps = {k: v for k, v in sorted(self.fps.items(),
                   key=lambda item: item[1], reverse=True)}
            fns = {k: v for k, v in sorted(self.fns.items(),
                   key=lambda item: item[1], reverse=True)}
            tps = {k: v for k, v in sorted(self.tps.items(),
                   key=lambda item: item[1], reverse=True)}

            # F1 per concept
            for cui in tps.keys():
                prec = tps[cui] / (tps.get(cui, 0) + fps.get(cui, 0))
                rec = tps[cui] / (tps.get(cui, 0) + fns.get(cui, 0))
                f1 = 2 * (prec * rec) / (prec + rec)
                self.cui_prec[cui] = prec
                self.cui_rec[cui] = rec
                self.cui_f1[cui] = f1

            # Get top 10
            pr_fps = [(self._get_pref_name(cui),
                      cui, fps[cui]) for cui in list(fps.keys())[0:10]]
            pr_fns = [(self._get_pref_name(cui),
                       cui, fns[cui]) for cui in list(fns.keys())[0:10]]
            pr_tps = [(self._get_pref_name(cui),
                       cui, tps[cui]) for cui in list(tps.keys())[0:10]]

            if do_print:
                print("\n\nFalse Positives\n")
                for one in pr_fps:
                    print("{:70} - {:20} - {:10}".format(str(one[0])[0:69],
                                                         str(one[1])[0:19],
                                                         one[2]))
                print("\n\nFalse Negatives\n")
                for one in pr_fns:
                    print("{:70} - {:20} - {:10}".format(str(one[0])[0:69],
                                                         str(one[1])[0:19],
                                                         one[2]))
                print("\n\nTrue Positives\n")
                for one in pr_tps:
                    print("{:70} - {:20} - {:10}".format(str(one[0])[0:69],
                                                         str(one[1])[0:19],
                                                         one[2]))
                print("*" * 110 + "\n")

        except Exception:
            traceback.print_exc()

    def _empty(self, cui: str) -> CUIInfo:
        return get_new_cui_info(
            cui=cui, preferred_name=cui, names=set((cui, )))

    def _get_or_empty(self, cui: str) -> CUIInfo:
        return self.cui2info.get(cui, self._empty(cui))

    def _get_pref_name(self, cui: str) -> str:
        info = self._get_or_empty(cui)
        return info['preferred_name'] or list(info['names'])[0]

    def unwrap(self) -> tuple[
        dict[str, int], dict[str, int], dict[str, int],
        dict[str, float], dict[str, float], dict[str, float],
        dict[str, int], dict
    ]:
        return (self.fps, self.fns, self.tps,
                self.cui_prec, self.cui_rec, self.cui_f1,
                self.cui_counts, self.examples)

    @classmethod
    def from_cat(cls, cat: CAT,
                 use_project_filters: bool = False,
                 use_overlaps: bool = False,
                 #  use_cui_doc_limit: bool = False,
                 #  use_groups: bool = False,
                 extra_cui_filter: Optional[set[str]] = None
                 ) -> 'StatsBuilder':
        """Get the stats builder from a model pack and some extra information.

        Args:
            cat (CAT):
                The model pack.
            use_project_filters (bool, optional):
                Whether to use per project filters. Defaults to False.
            use_overlaps (bool, optional):
                Whether to allow overlaps. Defaults to False.
            extra_cui_filter (Optional[set[str]], optional):
                Extra CUI filter. Defaults to None.

        Returns:
            StatsBuilder: The stats builder.
        """
        return StatsBuilder(addl_info=cat.cdb.addl_info,
                            filters=cat.config.components.linking.filters,
                            doc_getter=cat.__call__,
                            # cui2group=cat.cdb.addl_info['cui2group'],
                            # cui2preferred_name=cat.cdb.cui2preferred_name,
                            cui2info=cat.cdb.cui2info,
                            use_project_filters=use_project_filters,
                            use_overlaps=use_overlaps,
                            # use_cui_doc_limit=use_cui_doc_limit,
                            # use_groups=use_groups,
                            extra_cui_filter=extra_cui_filter)


def get_stats(cat: CAT,
              data: MedCATTrainerExport,
              epoch: int = 0,
              use_project_filters: bool = False,
              use_overlaps: bool = False,
              #   use_cui_doc_limit: bool = False,
              #   use_groups: bool = False,
              extra_cui_filter: Optional[set[str]] = None,
              do_print: bool = True) -> tuple[
        dict[str, int], dict[str, int], dict[str, int],
        dict[str, float], dict[str, float], dict[str, float],
        dict[str, int], dict
]:
    """TODO: Refactor and make nice
    Print metrics on a dataset (F1, P, R), it will also print the concepts
    that have the most FP,FN,TP.

    Args:
        cat: (CAT):
            The model pack.
        data (dict):
            The json object that we get from MedCATtrainer on export.
        epoch (int):
            Used during training, so we know what epoch is it.
        use_project_filters (bool):
            Each project in MedCATtrainer can have filters, do we want to
            respect those filters when calculating metrics.
        use_overlaps (bool):
            Allow overlapping entities, nearly always False as it is very
            difficult to annotate overlapping entities.
        use_cui_doc_limit (bool):
            If True the metrics for a CUI will be only calculated if that CUI
            appears in a document, in other words if the document was
            annotated for that CUI. Useful in very specific situations when
            during the annotation process the set of CUIs changed.
        use_groups (bool):
            If True concepts that have groups will be combined and stats will
            be reported on groups.
        extra_cui_filter(Optional[set]):
            This filter will be intersected with all other filters, or if all
            others are not set then only this one will be used.
        do_print (bool):
            Whether to print stats out. Defaults to True.

    Returns:
        fps (dict):
            False positives for each CUI.
        fns (dict):
            False negatives for each CUI.
        tps (dict):
            True positives for each CUI.
        cui_prec (dict):
            Precision for each CUI.
        cui_rec (dict):
            Recall for each CUI.
        cui_f1 (dict):
            F1 for each CUI.
        cui_counts (dict):
            Number of occurrence for each CUI.
        examples (dict):
            Examples for each of the fp, fn, tp.
            Format will be examples['fp']['cui'][<list_of_examples>].
    """
    builder = StatsBuilder.from_cat(cat,
                                    use_project_filters=use_project_filters,
                                    use_overlaps=use_overlaps,
                                    # use_cui_doc_limit=use_cui_doc_limit,
                                    # use_groups=use_groups,
                                    extra_cui_filter=extra_cui_filter)
    for pind, project in tqdm(enumerate(data['projects']),
                              desc="Stats project",
                              total=len(data['projects']),
                              leave=False):
        with project_filters(cat.config.components.linking.filters,
                             project,
                             builder.extra_cui_filter,
                             builder.use_project_filters):
            builder.process_project(project)
    # this is the part that prints out the stats
    builder.finalise_report(epoch, do_print=do_print)
    return builder.unwrap()



from ast import literal_eval
from typing import Any, Iterable, Union, Optional, cast
import random
import logging

import pandas
from torch.utils.data import Dataset
import torch

from medcat.cdb import CDB
from medcat.config.config import Config as CoreConfig
from medcat.config.config_rel_cat import ConfigRelCAT
from medcat.components.addons.relation_extraction.tokenizer import (
    BaseTokenizerWrapper)
from medcat.tokenizing.tokens import MutableEntity, MutableDocument
from medcat.data.mctexport import MedCATTrainerExportDocument

logger = logging.getLogger(__name__)


class RelData(Dataset):

    name = "rel_dataset"

    def __init__(self, tokenizer: BaseTokenizerWrapper,
                 config: ConfigRelCAT, cdb: CDB = CDB(CoreConfig())):
        """Use this class to create a dataset for relation annotations from
        CSV exports, MedCAT exports or Spacy Documents (assuming the documents
        got generated by MedCAT, if they did not then please set the required
        parameters manually to match MedCAT output, see
        /medcat/cat.py#_get_entity)

        If you are using this to create relations from CSV it is assumed that
        your entities/concepts of interest are surrounded by the special
        tokens, see create_base_relations_from_csv doc.

        Args:
            tokenizer (BaseTokenizerWrapper):
                tokenizer used to generate token ids from input text
            config (ConfigRelCAT): same config used in RelCAT
            cdb (CDB): Optional, used to add concept ids and types to
                detected ents,  useful when creating datasets from MedCAT
                output. Defaults to CDB().
        """

        self.cdb: CDB = cdb
        self.config: ConfigRelCAT = config
        self.tokenizer: BaseTokenizerWrapper = tokenizer
        self.dataset: dict[Any, Any] = {}

        logger.setLevel(self.config.general.log_level)

    def generate_base_relations(self, docs: Iterable[MutableDocument]
                                ) -> list[dict]:
        """ Util function, should be used if you want to train from spacy docs

        Args:
            docs (Iterable[MutableDocument]):
                Generate relations from Spacy CAT docs.

        Returns:
            output_relations: list[dict] : []
                "output_relations": relation_instances,
                # NOTE: see create_base_relations_from_doc/csv
                        for data columns
                "nclasses": self.config.model.padding_idx # dummy class
                "labels2idx": {},
                "idx2label": {}}
                ]
        """

        output_relations = []
        for doc_id, doc in enumerate(docs):
            output_relations.append(
                self.create_base_relations_from_doc(doc, doc_id=str(doc_id),))

        return output_relations

    def create_base_relations_from_csv(self, csv_path: str,
                                       keep_source_text: bool = False):
        """
            Assumes the columns are as follows
            ["relation_token_span_ids",
             "ent1_ent2_start", "ent1", "ent2", "label",
             "label_id", "ent1_type", "ent2_type",
             "ent1_id", "ent2_id", "ent1_cui", "ent2_cui", "doc_id", "sents"],
            last column is the actual source text.

            The entities inside the text MUST be annotated with special
            tokens i.e:
                ...text..[s1] first ent [e1].....[s2] second ent [e2]........
            You have to store the start position, aka index position of token
            [e1] and also of token [e2] in the (ent1_ent2_start) column.

        Args:
            csv_path (str): Path to csv file, must have specific columns,
                tab separated.
            keep_source_text (bool): If the text clumn should be retained in
                the 'sents' df column, used for debugging or creating
                custom datasets.

        Returns:
            dict : {
                "output_relations": relation_instances,
                # NOTE: see create_base_relations_from_doc/csv
                        for data columns
                "nclasses": self.config.model.padding_idx, # dummy class
                "labels2idx": {},
                "idx2label": {}}
            }
        """

        df = pandas.read_csv(csv_path, index_col=False,
                             encoding='utf-8', sep="\t")

        tmp_col_rel_token_col = df.pop("relation_token_span_ids")

        df.insert(0, "relation_token_span_ids", tmp_col_rel_token_col)

        text_cols = ["sents", "text"]

        df["ent1_ent2_start"] = df["ent1_ent2_start"].apply(
            lambda x: literal_eval(str(x)))

        for col in text_cols:
            if col in df.columns:
                out_rels = []
                for row_idx in range(len(df[col])):
                    _text = df.iloc[row_idx][col]
                    _ent1_ent2_start = df.iloc[row_idx]["ent1_ent2_start"]
                    _rels = self.create_base_relations_from_doc(
                        _text, doc_id=str(row_idx),
                        ent1_ent2_tokens_start_pos=_ent1_ent2_start,)
                    out_rels.append(_rels)

                rows_to_remove = []
                for row_idx in range(len(out_rels)):
                    if len(out_rels[row_idx]["output_relations"]) < 1:
                        rows_to_remove.append(row_idx)

                relation_token_span_ids = []
                out_ent1_ent2_starts = []

                for rel in out_rels:
                    if len(rel["output_relations"]) > 0:
                        relation_token_span_ids.append(
                            rel["output_relations"][0][0])
                        out_ent1_ent2_starts.append(
                            rel["output_relations"][0][1])
                    else:
                        relation_token_span_ids.append([])
                        out_ent1_ent2_starts.append([])

                df["label"] = [i.strip() for i in df["label"]]

                df["relation_token_span_ids"] = relation_token_span_ids
                df["ent1_ent2_start"] = out_ent1_ent2_starts

                df = df.drop(index=rows_to_remove)
                text_col = df.pop(col)
                df = df.assign(col=text_col)
                if keep_source_text:
                    df = df.assign(col=text_col)
                break

        nclasses, labels2idx, idx2label = RelData.get_labels(
            df["label"], self.config)

        output_relations = df.values.tolist()

        logger.info(
            "CSV dataset | No. of relations detected: %d "
            "| from : %s | nclasses: %d | idx2label: %s",
            len(output_relations), csv_path, nclasses,
            str(idx2label))

        logger.info("Samples per class: ")
        for label_num in list(idx2label.keys()):
            sample_count = 0
            for output_relation in output_relations:
                if idx2label[label_num] == output_relation[4]:
                    sample_count += 1
            logger.info(" label: %s | samples: %d",
                        idx2label[label_num], sample_count)

        # replace/update label_id with actual detected label number
        for idx in range(len(output_relations)):
            output_relations[idx][5] = int(
                labels2idx[output_relations[idx][4]])

        return {"output_relations": output_relations, "nclasses": nclasses,
                "labels2idx": labels2idx, "idx2label": idx2label}

    def _create_relation_validation(self,
                                    text: Union[str, MutableDocument],
                                    doc_id: str,
                                    tokenized_text_data: dict[str, Any],
                                    ent1_start_char_pos: int,
                                    ent2_start_char_pos: int,
                                    ent1_end_char_pos: int,
                                    ent2_end_char_pos: int,
                                    ent1_token_start_pos: int = -1,
                                    ent2_token_start_pos: int = -1,
                                    ent1_token_end_pos: int = -1,
                                    ent2_token_end_pos: int = -1,
                                    is_spacy_doc: bool = False,
                                    is_mct_export: bool = False,
                                    ) -> list:
        """
            This function checks if the relation is actually valid by distance
            criteria, TUIs and so on. Has diffierent handling cases for text,
            spacy docs and MCT exports.

        Args:
            text (str): doc text
            doc_id (str): doc id
            tokenized_text_data (dict[str, Any]): tokenized text
            ent1_start_char_pos (int): ent1 start char pos
            ent2_start_char_pos (int): ent2 start char pos
            ent1_end_char_pos (int): ent1 end char pos
            ent2_end_char_pos (int): ent2 end char pos
            ent1_token_start_pos (int): ent1_token_start_pos. Defaults to -1.
            ent2_token_start_pos (int): ent2_token_start_pos. Defaults to -1.
            ent1_token_end_pos (int): ent1_token_end_pos. Defaults to -1.
            ent2_token_end_pos (int): ent2_token_end_pos. Defaults to -1.
            is_spacy_doc (bool): checks if doc is spacy docs.
                Defaults to False.
            is_mct_export (bool): chekcs if doc is a mct export.
                Defaults to False.

        Returns:
            list: row containing rel data ["relation_token_span_ids",
                "ent1_ent2_start", "ent1", "ent2", "label", "label_id",
                "ent1_type", "ent2_type", "ent1_id", "ent2_id", "ent1_cui",
                "ent2_cui", "doc_id", "sents"]
        """

        text_length: int = len(text)

        doc_token_length: int = len(tokenized_text_data["tokens"])

        tmp_doc_text = text

        ent1_token: Union[str, MutableEntity] = tmp_doc_text[
            ent1_start_char_pos: ent1_end_char_pos]
        ent2_token: Union[str, MutableEntity] = tmp_doc_text[
            ent2_start_char_pos: ent2_end_char_pos]

        if (abs(ent2_start_char_pos - ent1_start_char_pos
                ) <= self.config.general.window_size and
                ent1_token != ent2_token):

            ent1_left_ent_context_token_pos_end = (
                ent1_token_start_pos - self.config.general.cntx_left)
            left_context_start_char_pos = 0

            if ent1_left_ent_context_token_pos_end < 0:
                ent1_left_ent_context_token_pos_end = 0
            else:
                left_context_start_char_pos = (
                    tokenized_text_data["offset_mapping"][
                        ent1_left_ent_context_token_pos_end][0])

            ent2_right_ent_context_token_pos_end = (
                ent2_token_end_pos + self.config.general.cntx_right)
            right_context_end_char_pos = text_length

            # get correct position, don't get last token as it
            # can be the [SEP] or [EOS] token.
            if ent2_right_ent_context_token_pos_end >= (doc_token_length - 1):
                ent2_right_ent_context_token_pos_end = doc_token_length - 2
            else:
                right_context_end_char_pos = (
                    tokenized_text_data["offset_mapping"][
                        ent2_right_ent_context_token_pos_end][1])

            if left_context_start_char_pos > right_context_end_char_pos:
                tmp = right_context_end_char_pos
                right_context_end_char_pos = left_context_start_char_pos
                left_context_start_char_pos = tmp

            if is_spacy_doc or is_mct_export:
                tmp_doc_text = text
                _pre_e1 = tmp_doc_text[0: (ent1_start_char_pos)]
                _e1_s2 = (
                    tmp_doc_text[ent1_end_char_pos: ent2_start_char_pos - 1])
                _e2_end = tmp_doc_text[ent2_end_char_pos + 1: text_length]
                ent2_token_end_pos = (ent2_token_end_pos + 2)

                annotation_token_text = (
                    self.tokenizer.hf_tokenizers.convert_ids_to_tokens(
                        self.config.general.annotation_schema_tag_ids))

                tmp_doc_text = (
                    str(_pre_e1) + " " +
                    annotation_token_text[0] + " " +
                    str(ent1_token) + " " +
                    annotation_token_text[1] + " " + str(_e1_s2) + " " +
                    annotation_token_text[2] + " " + str(ent2_token) + " " +
                    annotation_token_text[3] + " " + str(_e2_end)
                )

                ann_tag_token_len = len(annotation_token_text[0])

                _left_context_start_char_pos = (  # - 2 spaces
                    left_context_start_char_pos - ann_tag_token_len - 2)
                left_context_start_char_pos = max(
                    _left_context_start_char_pos, 0)

                _right_context_start_end_pos = (  # 8 for spces
                    right_context_end_char_pos + (ann_tag_token_len * 4) + 8)
                right_context_end_char_pos = (
                    len(tmp_doc_text) + 1 if
                    right_context_end_char_pos >= len(tmp_doc_text) or
                    _right_context_start_end_pos >= len(tmp_doc_text)
                    else _right_context_start_end_pos)

                # reassign the new text with added tags
                text_length = len(tmp_doc_text)

            # may lead to problems down the line if truncation=False,
            # if it is True we enforce the max 512 token length sentence
            # take care when using window_size > 300, we want to make sure
            # both entities are included at least... otherwise the relation
            # is considered invalid
            window_tokenizer_data = self.tokenizer(
                tmp_doc_text[
                    left_context_start_char_pos:right_context_end_char_pos],
                truncation=True)

            if self.config.general.annotation_schema_tag_ids:
                try:
                    ent1_token_start_pos = \
                        window_tokenizer_data["input_ids"].index(
                            self.config.general.annotation_schema_tag_ids[0])
                    ent2_token_start_pos = \
                        window_tokenizer_data["input_ids"].index(
                            self.config.general.annotation_schema_tag_ids[2])
                    _ent1_token_end_pos = \
                        window_tokenizer_data["input_ids"].index(
                            self.config.general.annotation_schema_tag_ids[1])
                    _ent2_token_end_pos = \
                        window_tokenizer_data["input_ids"].index(
                            self.config.general.annotation_schema_tag_ids[3])
                    assert ent1_token_start_pos
                    assert ent2_token_start_pos
                    assert _ent1_token_end_pos
                    assert _ent2_token_end_pos
                except Exception as exception:
                    logger.error(
                        "document id : %s failed to process relation",
                        str(doc_id), exc_info=exception)
                    return []

            if not self.config.general.annotation_schema_tag_ids:
                # update token loc to match new selection
                ent2_token_start_pos = (
                    ent2_token_start_pos - ent1_token_start_pos)
                ent1_token_start_pos = (
                    self.config.general.cntx_left if
                    ent1_token_start_pos - self.config.general.cntx_left > 0
                    else ent1_token_start_pos)
                ent2_token_start_pos += ent1_token_start_pos

            ent1_ent2_new_start = (ent1_token_start_pos, ent2_token_start_pos)
            en1_start, en1_end = window_tokenizer_data[
                "offset_mapping"][ent1_token_start_pos]
            en2_start, en2_end = window_tokenizer_data[
                "offset_mapping"][ent2_token_start_pos]

            return [window_tokenizer_data["input_ids"], ent1_ent2_new_start,
                    ent1_token, ent2_token, "UNK",
                    self.config.model.padding_idx,
                    None, None, None, None, None, None, doc_id, "",
                    en1_start, en1_end, en2_start, en2_end]
        return []

    def _get_token_type_and_start_end(
            self, entity: MutableEntity,
            doc_length_tokens: int, tokenized_text_data: dict[str, Any]
            ) -> tuple[list[str], tuple[int, int], tuple[int, int]]:
        ci = self.cdb.cui2info.get(entity.cui, None)
        type_ids = list(
            ci["type_ids"] if ci is not None else [])
        type_names = [
            self.cdb.addl_info["type_id2name"].get(tui, '')
            for tui in type_ids]

        start_char_pos = entity.base.start_char_index
        end_char_pos = entity.base.end_char_index

        token_start_pos = [
            i for i in range(0, doc_length_tokens)
            if start_char_pos in range(
                tokenized_text_data["offset_mapping"][i][0],
                tokenized_text_data["offset_mapping"][i][1] + 1)][0]
        token_end_pos = [
            i for i in range(0, doc_length_tokens)
            if end_char_pos in range(
                tokenized_text_data["offset_mapping"][i][0],
                tokenized_text_data["offset_mapping"][i][1] + 1)][0]
        return (type_names,
                (start_char_pos, end_char_pos),
                (token_start_pos, token_end_pos))

    def _create_base_relation_for_ents(
            self, doc_text: str, doc_id: str,
            ent1_token: MutableEntity, ent2_token: MutableEntity,
            tokenized_text_data: dict[str, Any],
            chars_to_exclude: str, doc_length_tokens: int
            ) -> Optional[list]:

        tmp_ent1 = ent1_token

        if (ent1_token.base.start_char_index >
                ent2_token.base.start_char_index):
            tmp_ent1 = ent1_token
            ent1_token = ent2_token
            ent2_token = tmp_ent1

        (ent1_types,
         (ent1_start_char_pos, ent1_end_char_pos),
         (ent1_token_start_pos,
          ent1_token_end_pos)) = self._get_token_type_and_start_end(
             ent1_token, doc_length_tokens, tokenized_text_data)

        (ent2_types,
         (ent2_start_char_pos, ent2_end_char_pos),
         (ent2_token_start_pos,
          ent2_token_end_pos)) = self._get_token_type_and_start_end(
             ent2_token, doc_length_tokens, tokenized_text_data)

        tkn1_str = str(ent1_token)
        tkn2_str = str(ent2_token)
        if (tkn2_str not in chars_to_exclude and
            tkn2_str not in
            self.tokenizer.hf_tokenizers.all_special_tokens and
                tkn1_str.strip() != tkn2_str.strip()):

            if self.config.general.relation_type_filter_pairs:
                for rel_pair in self.config.general.relation_type_filter_pairs:
                    if not (rel_pair[0] in ent1_types and
                            rel_pair[1] in ent2_types):
                        continue
                    return (
                        self._create_relation_validation(
                            text=doc_text,
                            doc_id=doc_id,
                            tokenized_text_data=tokenized_text_data,
                            ent1_start_char_pos=ent1_start_char_pos,
                            ent2_start_char_pos=ent2_start_char_pos,
                            ent1_end_char_pos=ent1_end_char_pos,
                            ent2_end_char_pos=ent2_end_char_pos,
                            ent1_token_start_pos=ent1_token_start_pos,
                            ent2_token_start_pos=ent2_token_start_pos,
                            ent1_token_end_pos=ent1_token_end_pos,
                            ent2_token_end_pos=ent2_token_end_pos,
                            is_spacy_doc=True
                        ))
            else:
                return (
                    self._create_relation_validation(
                        text=doc_text,
                        doc_id=doc_id,
                        tokenized_text_data=tokenized_text_data,
                        ent1_start_char_pos=ent1_start_char_pos,
                        ent2_start_char_pos=ent2_start_char_pos,
                        ent1_end_char_pos=ent1_end_char_pos,
                        ent2_end_char_pos=ent2_end_char_pos,
                        ent1_token_start_pos=ent1_token_start_pos,
                        ent2_token_start_pos=ent2_token_start_pos,
                        ent1_token_end_pos=ent1_token_end_pos,
                        ent2_token_end_pos=ent2_token_end_pos,
                        is_spacy_doc=True
                    ))
        return None

    def _create_base_relations_from_mutable_doc(
            self, doc: MutableDocument, doc_text: str, doc_id: str,
            tokenized_text_data: dict[str, Any],
            doc_length_tokens: int, chars_to_exclude: str) -> list[list]:
        _ents = doc.linked_ents if len(doc.linked_ents) > 0 else doc.ner_ents

        relation_instances: list[list] = []

        # last two can be a pair
        for ent1_idx in range(0, len(_ents) - 2):
            ent1_token = _ents[ent1_idx]

            if ((bt := ent1_token.base.text) not in chars_to_exclude and
                    bt not in
                    self.tokenizer.hf_tokenizers.all_special_tokens):
                for ent2_idx in range((ent1_idx + 1), len(_ents) - 1):
                    relation = self._create_base_relation_for_ents(
                        doc_text, doc_id, ent1_token, _ents[ent2_idx],
                        tokenized_text_data, chars_to_exclude,
                        doc_length_tokens)
                    if relation is not None:
                        relation_instances.append(relation)
        return relation_instances

    def create_base_relations_from_doc(
            self, doc: Union[MutableDocument, str], doc_id: str,
            ent1_ent2_tokens_start_pos: Union[list, tuple] = (-1, -1)) -> dict:
        """Creates a list of tuples based on pairs of entities detected
        (relation, ent1, ent2) for one spacy document or text string.

        Args:
            doc (Union[MutableDocument, str]): SpacyDoc or string of text,
                each will get handled slightly differently
            doc_id (str): Document id
            ent1_ent2_tokens_start_pos (Union[list, tuple], optional):
                Start of [s1][s2] tokens, if left default
                we assume we are dealing with a SpacyDoc.
                Defaults to (-1, -1).

        Returns:
                dict : {
                    # NOTE: see create_base_relations_from_doc/csv
                    #       for data columns
                    "output_relations": relation_instances,
                    "nclasses": self.config.model.padding_idx,  # dummy class
                    "labels2idx": {},
                    "idx2label": {}}
                }
        """

        (_ent1_start_tkn_id, _ent1_end_tkn_id,
         _ent2_start_tkn_id, _ent2_end_tkn_id) = 0, 0, 0, 0

        chars_to_exclude = ":!@#$%^&*()-+?_=.,;<>/[]{}"

        if self.config.general.annotation_schema_tag_ids:
            # we assume that ent1 start token is pos 0 and ent2 start token is
            # pos 2
            # e.g: [s1], [e1], [s2], [e2]
            _ent1_start_tkn_id = (
                self.config.general.annotation_schema_tag_ids[0])
            _ent1_end_tkn_id = self.config.general.annotation_schema_tag_ids[1]
            _ent2_start_tkn_id = (
                self.config.general.annotation_schema_tag_ids[2])
            _ent2_end_tkn_id = self.config.general.annotation_schema_tag_ids[3]

        relation_instances = []

        tokenized_text_data = None

        if isinstance(doc, str):
            doc_text = doc
        else:
            doc_text = doc.base.text

        tokenized_text_data = cast(dict[str, Any],
                                   self.tokenizer(doc_text, truncation=False))

        doc_length_tokens = len(tokenized_text_data["tokens"])

        if ent1_ent2_tokens_start_pos != (-1, -1) and isinstance(doc, str):
            ent1_token_start_pos = tokenized_text_data[
                "input_ids"].index(_ent1_start_tkn_id)
            ent2_token_start_pos = tokenized_text_data[
                "input_ids"].index(_ent2_start_tkn_id)
            ent1_token_end_pos = tokenized_text_data[
                "input_ids"].index(_ent1_end_tkn_id)
            ent2_token_end_pos = tokenized_text_data[
                "input_ids"].index(_ent2_end_tkn_id)

            ent1_start_char_pos, ent1_end_char_pos = tokenized_text_data[
                "offset_mapping"][ent1_token_start_pos]
            ent2_start_char_pos, ent2_end_char_pos = tokenized_text_data[
                "offset_mapping"][ent2_token_start_pos]

            relation_instances.append(self._create_relation_validation(
                text=doc_text,
                doc_id=doc_id,
                tokenized_text_data=tokenized_text_data,
                ent1_start_char_pos=ent1_start_char_pos,
                ent2_start_char_pos=ent2_start_char_pos,
                ent1_end_char_pos=ent1_end_char_pos,
                ent2_end_char_pos=ent2_end_char_pos,
                ent1_token_start_pos=ent1_token_start_pos,
                ent2_token_start_pos=ent2_token_start_pos,
                ent1_token_end_pos=ent1_token_end_pos,
                ent2_token_end_pos=ent2_token_end_pos
                ))
        elif not isinstance(doc, str):
            relation_instances.extend(
                self._create_base_relations_from_mutable_doc(
                    doc, doc_text, doc_id, tokenized_text_data,
                    doc_length_tokens, chars_to_exclude))

        # remove duplicates by using ent1_ent2_start_pos
        dupe_ent1_ent2_start = []

        _new_rel_instances = []
        for rel in relation_instances:
            if rel != []:
                if rel[1] not in dupe_ent1_ent2_start:
                    dupe_ent1_ent2_start.append(rel[1])
                    _new_rel_instances.append(rel)
                else:
                    logger.debug("removing duplicate relation" + str(rel[1]))

        # cleanup
        relation_instances = _new_rel_instances

        return {
            "output_relations": relation_instances,
            "nclasses": self.config.model.padding_idx,
            "labels2idx": {}, "idx2label": {}
        }

    def _create_relations_for_doc(
            self, document: MedCATTrainerExportDocument,
            data: dict,
            ) -> list[list]:
        doc_text: str = str(document["text"])
        doc_id: str = str(document["id"])

        if len(doc_text) == 0:
            return []
        annotations = document["annotations"]
        # these are not defined in the definition
        relations = document["relations"]  # type: ignore

        if self.config.general.lowercase:
            doc_text = doc_text.lower()

        tokenizer_text_data = self.tokenizer(
            doc_text, truncation=False)

        doc_token_length = len(tokenizer_text_data["tokens"])

        relation_instances = []
        ann_ids_from_relations = []

        ann_ids_ents: dict[Any, Any] = {}

        _other_relations_subset = []

        # this section creates 'Other' class relations
        # based on validated annotations
        for ent1_idx, ent1_ann in enumerate(annotations):
            ann_id = ent1_ann["id"]
            ann_ids_ents[ann_id] = {}
            ann_ids_ents[ann_id]["cui"] = ent1_ann["cui"]
            ci1 = self.cdb.cui2info.get(ent1_ann["cui"], None)
            ann_ids_ents[ann_id]["type_ids"] = list(
                ci1["type_ids"] if ci1 is not None else [])
            ann_ids_ents[ann_id]["types"] = [
                self.cdb.addl_info['type_id2name'].get(tui, '')
                for tui in ann_ids_ents[ann_id]['type_ids']]

            ent1_types = ann_ids_ents[ann_id]["types"]

            if self.config.general.create_addl_rels:
                for _, ent2_ann in enumerate(
                        annotations[ent1_idx + 1:]):
                    ci2 = self.cdb.cui2info.get(
                        ent2_ann["cui"], None)
                    ent2_types = list(
                        ci2["type_ids"] if ci2 is not None else [])

                    if not (ent1_ann["validated"] and ent2_ann[
                            "validated"]):
                        continue
                    _relation_type = "Other"

                    # create new Other subclass class if enabled
                    if self.config.general.create_addl_rels_by_type:
                        _relation_type = (
                            "Other" + ent1_types[0] + "-" + ent2_types[0])

                    _other_relations_subset.append({
                        "start_entity": ent1_ann["id"],
                        "start_entity_cui": ent1_ann["cui"],
                        "start_entity_value": ent1_ann["value"],
                        "start_entity_start_idx": ent1_ann["start"],
                        "start_entity_end_idx": ent1_ann["end"],
                        "end_entity": ent2_ann["id"],
                        "end_entity_cui": ent2_ann["cui"],
                        "end_entity_value": ent2_ann["value"],
                        "end_entity_start_idx": ent2_ann["start"],
                        "end_entity_end_idx": ent2_ann["end"],
                        "relation": _relation_type,
                        "validated": True
                    })

        non_rel_sample_size_limit = int(int(
            self.config.general.addl_rels_max_sample_size) /
            len(data['projects']))

        if (non_rel_sample_size_limit > 0 and
                len(_other_relations_subset) > 0):
            random.shuffle(_other_relations_subset)
            _other_relations_subset = _other_relations_subset[
                0:non_rel_sample_size_limit]

        relations.extend(_other_relations_subset)

        for relation in relations:
            ann_start_start_pos = relation['start_entity_start_idx']
            ann_start_end_pos = relation["start_entity_end_idx"]

            ann_end_start_pos = relation['end_entity_start_idx']
            ann_end_end_pos = relation["end_entity_end_idx"]

            start_entity_value = relation['start_entity_value']
            end_entity_value = relation['end_entity_value']

            start_entity_id = relation['start_entity']
            end_entity_id = relation['end_entity']

            start_entity_types = ann_ids_ents[start_entity_id]['types']
            end_entity_types = ann_ids_ents[end_entity_id]['types']
            start_entity_cui = ann_ids_ents[start_entity_id]['cui']
            end_entity_cui = ann_ids_ents[end_entity_id]['cui']

            # if somehow the annotations belong to the same relation
            # but make sense in reverse
            if ann_start_start_pos > ann_end_start_pos:
                ann_end_start_pos = relation['start_entity_start_idx']
                ann_end_end_pos = relation['start_entity_end_idx']

                ann_start_start_pos = relation['end_entity_start_idx']
                ann_start_end_pos = relation['end_entity_end_idx']

                end_entity_value = relation['start_entity_value']
                start_entity_value = relation['end_entity_value']

                end_entity_cui = ann_ids_ents[start_entity_id]['cui']
                start_entity_cui = ann_ids_ents[end_entity_id]['cui']

                end_entity_types = ann_ids_ents[start_entity_id]['types']
                start_entity_types = ann_ids_ents[end_entity_id]['types']

                # switch ids last
                start_entity_id = relation['end_entity']
                end_entity_id = relation['start_entity']

            for ent1type, ent2type in enumerate(
                    self.config.general.relation_type_filter_pairs
                    ):
                if (ent1type not in start_entity_types and
                        ent2type not in end_entity_types):
                    # TODO: Does this make any sense? I don't think so...
                    #       It's probably meant to continue on the wider loop
                    continue

            ann_ids_from_relations.extend([
                start_entity_id, end_entity_id])
            relation_label = relation['relation'].strip()

            try:

                ent1_token_start_pos = [
                    i for i in range(0, doc_token_length)
                    if ann_start_start_pos in range(
                        tokenizer_text_data["offset_mapping"][i][0],
                        tokenizer_text_data["offset_mapping"][i][1] + 1)][0]

                ent2_token_start_pos = [
                    i for i in range(0, doc_token_length)
                    if ann_end_start_pos in range(
                        tokenizer_text_data["offset_mapping"][i][0],
                        tokenizer_text_data["offset_mapping"][i][1] + 1)][0]

                ent1_token_end_pos = [
                    i for i in range(0, doc_token_length)
                    if ann_start_end_pos in range(
                        tokenizer_text_data["offset_mapping"][i][0],
                        tokenizer_text_data["offset_mapping"][i][1] + 1)][0]

                ent2_token_end_pos = [
                    i for i in range(0, doc_token_length)
                    if ann_end_end_pos in range(
                        tokenizer_text_data["offset_mapping"][i][0],
                        tokenizer_text_data["offset_mapping"][i][1] + 1)][0]
                assert ent1_token_start_pos
                assert ent2_token_start_pos
                assert ent1_token_end_pos
                assert ent2_token_end_pos
            except Exception as e:
                logger.info(
                    "document id: %s failed to process relation",
                    str(doc_id), exc_info=e)
                raise e

            if (start_entity_id != end_entity_id and relation.get(
                "validated", True) and
                start_entity_value not in
                self.tokenizer.hf_tokenizers.all_special_tokens
                    and end_entity_value not in
                    self.tokenizer.hf_tokenizers.all_special_tokens):
                final_relation = self._create_relation_validation(
                    text=doc_text,
                    doc_id=doc_id,
                    tokenized_text_data=tokenizer_text_data,
                    ent1_start_char_pos=ann_start_start_pos,
                    ent2_start_char_pos=ann_end_start_pos,
                    ent1_end_char_pos=ann_start_end_pos,
                    ent2_end_char_pos=ann_end_end_pos,
                    ent1_token_start_pos=ent1_token_start_pos,
                    ent2_token_start_pos=ent2_token_start_pos,
                    ent1_token_end_pos=ent1_token_end_pos,
                    ent2_token_end_pos=ent2_token_end_pos,
                    is_mct_export=True
                )

                if len(final_relation) > 0:
                    final_relation[4] = relation_label
                    final_relation[6] = start_entity_types
                    final_relation[7] = end_entity_types
                    final_relation[8] = start_entity_id
                    final_relation[9] = end_entity_id
                    final_relation[10] = start_entity_cui
                    final_relation[11] = end_entity_cui

                    relation_instances.append(final_relation)
        return relation_instances

    def create_relations_from_export(self, data: dict):
        """
            Args:
                data (dict):
                    MedCAT Export data.

            Returns:
                dict : {
                    # NOTE: see create_base_relations_from_doc/csv
                            for data columns
                    "output_relations": relation_instances,
                    "nclasses": self.config.model.padding_idx,  # dummy class
                    "labels2idx": {},
                    "idx2label": {}}
                }
        """

        output_relations = []

        for project in data["projects"]:
            for _doc_id, document in enumerate(project["documents"]):
                relation_instances = self._create_relations_for_doc(
                    document, data)
                output_relations.extend(relation_instances)

        all_relation_labels = [relation[4] for relation in output_relations]

        nclasses, labels2idx, idx2label = self.get_labels(
            all_relation_labels, self.config)

        # replace label_id with actual detected label number
        for idx in range(len(output_relations)):
            output_relations[idx][5] = int(
                labels2idx[output_relations[idx][4]])

        logger.info("MCT export dataset | nclasses: %d | idx2label: %s",
                    nclasses, str(idx2label))
        logger.info("Samples per class: ")

        logger.error(str(idx2label))

        for label_num in list(idx2label.keys()):
            sample_count = 0
            for output_relation in output_relations:
                if idx2label[label_num] == output_relation[4]:
                    sample_count += 1
            logger.info(
                " label: %s | samples: %s",
                idx2label[label_num], str(sample_count))

        return {"output_relations": output_relations, "nclasses": nclasses,
                "labels2idx": labels2idx, "idx2label": idx2label}

    @classmethod
    def get_labels(cls, relation_labels: list[str], config: ConfigRelCAT
                   ) -> tuple[int, dict[str, int], dict[int, str]]:
        """This is used to update labels in config with unencountered
        classes/labels ( if any are encountered during training).

        Args:
            relation_labels (list[str]): new labels to add
            config (ConfigRelCAT): config

        Returns:
            tuple[int, dict[str, int], dict[int, str]]: label count,
                labesl2idx mapping, idx2labels mapping
        """
        curr_class_id = 0

        config_labels2idx: dict[str, int] = config.general.labels2idx
        config_idx2labels: dict[int, str] = config.general.idx2labels

        relation_labels = [relation_label.strip()
                           for relation_label in relation_labels]

        for relation_label in set(relation_labels):
            if relation_label not in config_labels2idx.keys():
                while curr_class_id in [
                        int(label_idx) for
                        label_idx in config_idx2labels.keys()]:
                    curr_class_id += 1
                config_labels2idx[relation_label] = int(curr_class_id)
                config_idx2labels[int(curr_class_id)] = relation_label

        return (len(config_labels2idx.keys()), config_labels2idx,
                config_idx2labels,)

    def __len__(self) -> int:
        """
        Returns:
            int: num of rels records
        """
        return len(self.dataset['output_relations'])

    def __getitem__(self, idx: int) -> tuple[torch.LongTensor,
                                             torch.LongTensor,
                                             torch.LongTensor]:
        """

        Args:
            idx (int): index of item in the dataset dict

        Returns:
            tuple[torch.LongTensor, torch.LongTensor, torch.LongTensor]:
                Long tensors of the following the columns :
                    input_ids, ent1&ent2 token start pos idx, label_ids
        """

        return (
            torch.LongTensor(self.dataset['output_relations'][idx][0]),
            torch.LongTensor(self.dataset['output_relations'][idx][1]),
            torch.LongTensor([self.dataset['output_relations'][idx][5]])
        )
